import requests
from colorama import init, Fore, Style
import time
import json
import math
import numpy as np
from threed_optix.package_utils import vars as v
import pandas as pd
import threed_optix.package_utils.math as mu
import copy

def verify_response(response):
    '''
    This function checkes if the response is valid.
    '''

    #If the response status code is not in the valid response codes, raise an exception.
    if int(response.status_code) not in v.VALID_RESPONSE_CODES:
        raise Exception(f'Error: {response.status_code} - {response.text}')
    return None


#General Purpose
def _healthcheck():
    """
    Checks if the server is up and well.
    """
    url = f'{v.API_URL}/healthcheck'

    r = requests.get(url)
    verify_response(r)
    r = r.json()
    return (r['status'] == 'SUCCESS', r['message'])

def _get(
    endpoint: str,
    api_key: str,
    #payload: dict = {}
    ):
    """
    Sends a GET request to the endpoint with the API key.

    Args:
        endpoint (str): The endpoint to send the request to.
        api_key (str): The API key.

    Returns:
        tuple: The data and message from the response.
    """

    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key,
               'Content-Type': 'application/json',
               "sdk-version": v.VERSION,
               'is_new_session': v.is_new_session}

    if v.DEBUG:
        print(f'GET {url}')
        print(f'Headers: {headers}')

    r = requests.get(url, headers=headers)
    verify_response(r)
    r = r.json()

    if v.DEBUG:
        print(f'Response: {r}')

    # Return the data if the status is SUCCESS, otherwise return None, and the message.
    return (r['data'] if 'status' in r and r['status'] == 'SUCCESS' else None, r['message'])

def _put(endpoint: str, api_key: str, json_data: dict = None):
    """
    Sends a PUT request to the endpoint.

    Args:
        endpoint (str): The endpoint to send the request to.
        json_data (dict): The JSON payload for the request.
        api_key (str): The API key.

    Returns:
        tuple: The data and message from the response.
    """
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key,
               'Content-Type': 'application/json',
               "sdk-version": v.VERSION,
                'is_new_session': v.is_new_session}

    if v.DEBUG:
        print(f'PUT {url}')
        print(f'Headers: {headers}')
        print(f'Payload: {json_data}')

    r = requests.put(url, headers=headers, json=json_data)
    verify_response(r)
    r = r.json()

    if v.DEBUG:
        print(f'Response: {r}')

    # Return the data if the status is SUCCESS, otherwise return None, and the message.
    return (r['data'] if 'status' in r and r['status'] == 'SUCCESS' else None, r['message'])

def _set(endpoint, data, api_key):
    '''
    Sends a POST request to the endpoint with the data and the API key.
    tbh, I don't know why it's seperated from the _post function.

    Args:
        endpoint (str): The endpoint to send the request to.
        data (dict): The data to send.
        api_key (str): The API key.

    Returns:
        tuple: The data and message from the response.
    '''
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key,
               'Content-Type': 'application/json',
               "sdk-version": v.VERSION,
                'is_new_session': v.is_new_session}


    if v.DEBUG:
        print(f'SET {url}')
        print(f'Headers: {headers}')
        print(f'Payload: {data}')


    r = requests.post(url, headers=headers, json=data)

    verify_response(r)
    r = r.json()

    if v.DEBUG:
        print(f'Response: {r}')

    return ('status' in r and r['status'] == 'SUCCESS', r['message'])

def _post(endpoint, data, api_key):
    '''
    Sends a POST request to the endpoint with the data and the API key.

    Args:
        endpoint (str): The endpoint to send the request to.
        data (dict): The data to send.
        api_key (str): The API key.

    Returns:
        tuple: The status, message, and data from the response.
    '''
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key,
               'Content-Type': 'application/json',
               "sdk-version": v.VERSION,
                'is_new_session': v.is_new_session}


    if v.DEBUG:
        print(f'POST {url}')
        print(f'Headers: {headers}')
        print(f'Payload: {data}')

    r = requests.post(url, headers=headers, json=data)
    verify_response(r)
    r = r.json()

    if v.DEBUG:
        print(f'Response: {r}')

    return ('status' in r and r['status'] == 'SUCCESS', r['message'], r.get('data', None))

def _delete(endpoint, api_key):
    '''
    Sends a DELETE request to the endpoint with the API key.

    Args:
        endpoint (str): The endpoint to send the request to.
        api_key (str): The API key.

    Returns:
        tuple: The status and message from the response.
    '''
    url = f'{v.API_URL}/{endpoint}'
    headers = {'X-API-KEY': api_key,
               'Content-Type': 'application/json',
               "sdk-version": v.VERSION,
                'is_new_session': v.is_new_session}

    if v.DEBUG:
        print(f'DELETE {url}')
        print(f'Headers: {headers}')

    r = requests.delete(url, headers=headers)
    verify_response(r)
    r = r.json()

    if v.DEBUG:
        print(f'Response: {r}')

    return ('status' in r and r['status'] == 'SUCCESS', r['message'])

def _create_setup(parameters, api_key):
    '''
    This function accesses the API create setup endpoint.
    '''
    endpoint = v.POST_CREATE_SETUP_ENDPOINT
    response = _post(endpoint, parameters, api_key)
    if not response[0]:
        raise Exception(response[1])
    return response[2]

def _delete_part(setup_id, part_id, api_key):
    '''
    This function accesses the API delete part endpoint.
    '''
    endpoint = v.DELETE_PART_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    return _delete(endpoint, api_key)

def _add_part(setup_id, data, api_key):
    '''
    This function accesses the API add part endpoint.
    '''
    endpoint = v.POST_ADD_PART_ENDPOINT.format(setup_id=setup_id)
    response = _post(endpoint,data,api_key)
    if not response[0]:
        raise Exception(response[1])
    return response[2]

def _get_setups(api_key: str):
    """
    Returns the list of setup names and ids of the user.

    Args:
        api_key (str): The API key.

    Returns:
        list: A list of setup names and ids.
    """
    endpoint = v.GET_SETUPS_ENDPOINT
    response = _get(endpoint, api_key)

    return response

def _get_setup(
    setup_id: str,
    api_key: str
    ):
    """
    Returns the opt file for the specified setup id.

    Args:
        setup_id (str): The setup id.
        api_key (str): The API key.

    Returns:
        str: The opt file content.
    """

    endpoint = v.GET_SETUP_ENDPOINT.format(setup_id=setup_id)

    return _get(endpoint, api_key)

def _get_surface(api_key, setup_id, surface_id, part_id):
    '''
    This function fetches the information of a surface.
    '''
    endpoint = v.Endpoints.GET_SURFACE_DATA.format(setup_id=setup_id, surface_id=surface_id, part_id=part_id)
    return _get(endpoint, api_key)

def _create_part(parameters, api_key):
    '''
    This function accesses the API create part endpoint.
    '''
    endpoint = v.POST_CREATE_OPTICS_ENDPOINT
    response = _post(endpoint, parameters, api_key)
    if not response[0]:
        raise Exception(response[1])
    return response[2]

def _get_part(setup_id: str, part_id: str, api_key: str):
    '''
    This function fetches the information of a part.
    '''
    endpoint = v.GET_PART_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    part =  _get(endpoint, api_key)
    return part

def _change_part(setup_id, part_id, data, api_key):
    '''
    This function accesses the API change part endpoint.
    '''

    # Copy the data to avoid changing the original data.
    data_copy = copy.deepcopy(data)

    # Convert the rotation to radians.
    if data.get('pose'):
        data_copy['pose']['rotation'] =[mu.deg_to_rad(x) for x in data_copy['pose']['rotation']]
    endpoint = v.PUT_PART_ENDPOINT.format(setup_id=setup_id, part_id=part_id)
    r = _put(endpoint=endpoint, api_key=api_key, json_data=data_copy)
    return r

def _delete_analysis(setup_id, part_id, surface_id,analysis_id, api_key):
    '''
    This function accesses the API delete analysis endpoint.
    '''
    endpoint = v.Endpoints.DELETE_ANALYSIS.format(setup_id=setup_id,
                                                 part_id=part_id,
                                                 analysis_id=analysis_id,
                                                 surface_id = surface_id)
    return _delete(endpoint, api_key)

def _run_async(setup_id, api_key):
    '''
    Not suppoerted anymore.
    '''
    endpoint =  v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)
    json_data = {
        "gpu_type": v.GPU_TYPE,
    }
    r = _put(endpoint = endpoint, json_data=json_data, api_key= api_key)
    return r

def _run_batch(
    setup_id: str,
    configuration: dict,
    api_key: str):
    """
    Not supported anymore.
    Puts the batch run request.

    Args:
        setup_id (str): The setup id.
        configuration (dict): The batch configuration.
        api_key (str): The API key.
    """
    endpoint = v.PUT_BATCH_CHANGES_ENDPOINT.format(setup_id=setup_id)
    return _put(endpoint = endpoint, api_key= api_key, json_data = configuration)

def _run_simulation(setup_id, api_key, is_sync = True):
    '''
    This function accesses the API run simulation endpoint for propagation.
    '''
    endpoint =  v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)
    json_data = {
        "gpu_type": v.GPU_TYPE,
        "is_sync": is_sync,
    }
    r = _put(endpoint = endpoint, json_data=json_data, api_key= api_key)
    return r

def _add_analyses(setup_id, part_id, surface_id, data, api_key):
    '''
    This function accesses the API add analysis endpoint.
    '''
    endpoint = v.POST_ADD_ANALYSIS_ENDPOINT.format(setup_id=setup_id, part_id=part_id, surface_id = surface_id)
    return _post(endpoint = endpoint,data=data, api_key = api_key)

def _run_analysis(setup_id: str, api_key: str, analysis_id: str):
    '''
    This function accesses the API run simulation endpoint to run an analysis.
    '''
    endpoint = v.PUT_SIMULATION_ENDPOINT.format(setup_id=setup_id)

    json_data = {
        "gpu_type": v.GPU_TYPE,
        "is_sync": True,
        "analysis_id": analysis_id,
    }

    return _put(endpoint = endpoint, json_data = json_data, api_key = api_key)

def _change_scattering(setup_id, part_id, surface_id, data, api_key):
    '''
    This function accesses the API change scattering endpoint.
    '''
    endpoint = v.Endpoints.SCATTERING.format(setup_id=setup_id, part_id=part_id, surface_id=surface_id)
    return _put(endpoint=endpoint, api_key=api_key, json_data=data)

def _get_optics_data(api_key, db_id):
    '''
    This function fetches the optics data.
    '''
    endpoint = v.Endpoints.GET_OPTICS_DATA.format(number_id=db_id)
    return _get(endpoint, api_key)

def _change_cs_data(setup_id, part_id, data, api_key):
    '''
    This function accesses the API change cs endpoint.
    '''
    endpoint = v.Endpoints.CHANGE_CS.format(setup_id=setup_id, part_id=part_id)
    return _put(endpoint=endpoint, api_key=api_key, json_data=data)

def _ask(conversation, api_key):
    '''
    Not implemented yet.
    This function accesses the API ask endpoint.
    It allows the user to ask questions to the SDK Assistant
    '''
    endpoint = v.GET_ANSWER_ENDPOINT
    headers = {'X-API-KEY': api_key}
    params = {'threed_optix_key': api_key, "conversation": json.dumps(conversation)}
    r = requests.get(endpoint, headers=headers, params=params)
    verify_response(r)
    r = r.json()
    return r['answer']

def _set_api_url(url, are_you_sure=False):
    '''
    Private.
    This function sets the API URL.
    It should be used internally to switch the API URL to dev or release.
    '''

    if not are_you_sure:
        raise Exception(v.SET_API_URL_WARNING)

    print(v.SET_API_URL_WARNING)
    print('Previous API URL was', v.API_URL)
    print('Setting API URL to', url)
    v.API_URL = url

def _get_materials(api_key, material_name):
    '''
    This function fetches the materials data based on the material name.
    '''
    endpoint = v.Endpoints.GET_MATERIALS_ENDPOINT.format(material_name=material_name)
    data, message = _get(endpoint, api_key)
    if data is None:
        raise Exception(message)
    return data

def _set_ask_url(url, are_you_sure=False):
    '''
    Not implemented yet.
    This function sets the ASK URL.
    It should be used internally to switch the ASK URL to dev or release.
    '''
    if not are_you_sure:
        raise Exception(v.SET_API_URL_WARNING)

    print(v.SET_API_URL_WARNING)
    print('Previous ASK URL was', v.ASK_URL)
    print('Setting ASK URL to', url)
    v.ASK_URL = url

def _welcome():
    '''
    Private.
    This is the function that prints the welcome message, when the client object is created.
    '''
    init(autoreset=True)
    color = Fore.WHITE
    print('******************')
    for i, char in enumerate(v.WELCOME_MESSAGE):
        print(f"{color}{char}", end="")
        time.sleep(0.01)  # Adjust the delay for the desired speed

    print(Style.RESET_ALL)  # Reset the style after the rainbow effect
    print('******************')

def _print_getting_parts():
    '''
    Deprecated.
    This function used to print a message when we fetched the parts information of the setup.
    It is no longer needed since we are not fetching the parts information all at once anymore.
    '''
    init(autoreset=True)
    color = Fore.WHITE
    print('******************')
    for i, char in enumerate(v.GETTING_PARTS_MESSAGE):
        print(f"{color}{char}", end="")
        time.sleep(0.01)  # Adjust the delay for the desired speed

    print(Style.RESET_ALL)  # Reset the style after the rainbow effect
    print('******************')

def _map_ray_table(rays_url, maps_url):
    '''
    This function creates a pandas dataframe from the rays data and maps the indices to the actual values.
    Args:
        rays_url (str): The URL to the rays data, recieved from the simulation results.
        maps_url (str): The URL to the maps data, recieved from the simulation results.

    Returns:
        pandas.DataFrame: The rays data with the indices mapped to actual values.
    '''

    # Fetch the rays data and the maps data.
    rays_df = pd.read_csv(rays_url)

    # This is probably useless, but I'm documenting when I don't have time to examine changes.
    rays_df = rays_df.copy()

    # Fetch the maps data.
    maps_json = requests.get(maps_url).json()

    # Map the indices to the actual values.
    rays_df = rays_df.apply(lambda row: _map_ray_row(row, maps_json), axis=1)

    # Set the index to the idx column.
    rays_df.set_index('idx', inplace=True)
    return rays_df

def _map_ray_row(row, map):
    '''
    This is a helper function to map the indices to the actual values in the rays data per each row of the ray dataframe.
    '''

    # Map the surface from indices to the actual id.
    row['surface'] = map['surfaces'][int(row['hit_surface_idx'])]

    # Map the wavelength from indices to the actual value.
    row['wavelength'] = map['wavelengths'][int(row['wavelength_idx'])]

    # Map the light source from indices to the actual light source id.
    row['light_source'] = map['sources'][int(row['source_idx'])]

    # Drop the indices columns.
    del row['hit_surface_idx']
    del row['wavelength_idx']
    del row['source_idx']
    return row



from .client import ThreedOptixAPI, Client, Material
from .analyses import Analysis
from .simulations import Setup
from .parts import Part, Surface, LightSource, Detector, GLOBAL
from .utils import *
from . import package_utils
from . import optimize

# Allow users to easily access the enums
from .package_utils.vars import eOpticsSubtypes, Setups, Coating, SETUP_LABELS

__version__ = package_utils.vars.VERSION
GRATING_SUBTYPES = eOpticsSubtypes.GRATING_SUBTYPES
SETUP_LABELS = Setups.SETUP_LABELS

# Enter debug mode, which will print out pretty much every step of the way.
def debug(boolean = None):

    # If the user has specified a boolean, set the debug mode to that value.
    if boolean is not None:
        package_utils.vars.DEBUG = boolean

    # Otherwise, toggle the debug mode.
    else:
        package_utils.vars.DEBUG = not package_utils.vars.DEBUG



import zipfile
import requests
import random
import time
import json
import copy
import os
from typing import List, Dict, Union, Tuple
import pandas as pd
import io
import struct
import shutil

import pandas as pd
import plotly.express as px
import numpy as np
from io import BytesIO

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from plotly.graph_objs.layout import Colorscale
from functools import reduce

import threed_optix.package_utils.api as au
import threed_optix.package_utils.general as gu
import threed_optix.package_utils.vars as v
import threed_optix as tdo

class RayTable(pd.DataFrame):
    '''
    This is the class that represents the ray table.
    It is a pandas DataFrame with some additional attributes, mainly the setup object.
    '''
    def __init__(self, ray_table_url, maps_url, setup_object):
        '''
        Instanciates a new RayTable object.
        '''

        # Fetch the data from the URL
        df = au._map_ray_table(ray_table_url, maps_url)

        # Call the parent class constructor
        super().__init__(df)

        # Add the setup object as an attribute
        self.attrs['setup'] = setup_object
        return None

    @property
    def setup(self):
        return self.attrs['setup']

class Analysis:

    def __init__(self,
                 surface,
                 resolution: Union[Tuple[int, int], List[int]],
                 rays: Union[dict, int],
                 name: str,
                 version: int = 3,
                 fast: bool=False,
                 ):
        """
        Initializes a new instance of the Analysis class.

        Args:
            surface (Surface): The surface of the analysis.
            resolution (tuple): The resolution of the analysis surface in the form (x, y).
            rays (dict): A dictionary of lasers and the number of rays for each laser.
            name (str): The name of the analysis.
            fast (bool, optional): Not used in practice, Specifies if the analysis is fast or advanced. Defaults to False.

        Returns:
            tdo.Analysis: The created Analysis object.

        Raises:
            AssertionError: If the name or rays are not valid for 'fast' choice.
        """

        def verify(fast, name, rays, surface):
            '''
            Private.
            '''

            errors = []

            if not isinstance(surface, tdo.Surface):
                errors.append(f'surface must be of type Surface, got {surface.__class__.__name__}')

            if fast:
                #Check if the name of the analysis is valid for fast analysis
                if not name in v.FAST_ANALYSIS_NAMES:
                    errors.append(f"Valid names for fast analysis are {v.FAST_ANALYSIS_NAMES}")
                #Check if the number of rays is less than 200
                if not all([num <= 200 for num in rays.values()]):
                    errors.append(f'Number of rays must be less than 200 for fast analysis')
            else:
                #Check if the name of the analysis is valid for advanced analysis
                if name in v.DEPRICATED_ANALYSIS_NAMES:
                    non_polirized_name = name.replace(" Polarized", "")
                    errors.append(f"{name} analysis type name is deprecated. You can use {non_polirized_name} instead to use polarized analysis.")
                elif name not in v.VALID_ANALYSIS_NAMES:
                    errors.append(f"Valid names for advanced analysis are {v.VALID_ANALYSIS_NAMES}")



            #res is 2 floats tuple
            if len(resolution) != 2:
                errors.append(f'Resolution must be a tuple of 2 integers, got len {len(resolution)}')

            #Res in the acceptable range
            if not all([v.ANALYSIS_RES_RANGE[0] < num <= v.ANALYSIS_RES_RANGE[1] for num in resolution]):
                errors.append(f'Resolution must be between {v.ANALYSIS_RES_RANGE[0]} and {v.ANALYSIS_RES_RANGE[1]} for analysis')

            # Rays values are integers
            # Trust me, Don't check for ints, because 1e8 is a float, for example
            if not all([int(num) == num for num in rays.values()]):
                errors.append(f'Number of rays must be an integer')

            # If rays are a dictionary of lasers and number of rays
            if isinstance(rays, dict):
                # Rays keys are LightSource
                if not all([key.__class__.__name__ == 'LightSource' for key in rays.keys()]):
                    errors.append(f'Keys of rays must be of type LightSource')
                # Rays light sources are in the setup
                all_laser_ids_in_rays = self._laser_ids_in_rays(surface._part._setup, rays)
                if not all_laser_ids_in_rays:
                    errors.append('If rays is a dict, all lasers in the setup must be included in the rays dictionary')
                # Rays in the acceptable range
                if not all([v.ANALYSIS_RAYS_RANGE[0] <= num <= v.ANALYSIS_RAYS_RANGE[1] for num in rays.values()]):
                    errors.append(f'Number of rays must be between {v.ANALYSIS_RAYS_RANGE[0]} and {v.ANALYSIS_RAYS_RANGE[1]} for analysis')

            # If rays is an integer, indicating the number of rays for all lasers
            else:
                if v.ANALYSIS_RAYS_RANGE[0] <= rays <= v.ANALYSIS_RAYS_RANGE[1]:
                    errors.append(f'Number of rays must be between {v.ANALYSIS_RAYS_RANGE[0]} and {v.ANALYSIS_RAYS_RANGE[1]} for analysis')

            return errors

        # If rays is an integer, indicating the number of rays for all lasers
        if isinstance(rays, float) or isinstance(rays, int):
            rays = {laser: rays for laser in surface._part._setup.light_sources}

        if isinstance(resolution, (int, float)):
            resolution = (resolution, resolution)

        # Check if the arguments are valid
        errors = verify(fast=fast, name=name, rays=rays, surface = surface)
        if errors:
            raise AssertionError(v.argument_repair_message(errors))

        if not fast and name in (v.VALID_ANALYSIS_NAMES[0], v.VALID_ANALYSIS_NAMES[1], v.VALID_ANALYSIS_NAMES[2]):
            name = f"{name} Polarized"


        self._added = False
        self._urls = []
        self._fail_message = None
        self._raw_results = {}
        self.results = {}
        self.surface = surface
        self.name = name
        self.rays = rays
        self.resolution = resolution
        self.version = 3
        self.fast = fast

        # No need to generate id anymore, it is being set when added to the surface
        self.id = Analysis._generate_id()


    @classmethod
    def _new(cls, surface, resolution, num_rays, name, type, id):
        '''
        Private.
        Past analysis are stored within the setup.
        When the setup is fetched, the past analysis are created using this method.
        '''
        analysis = object.__new__(cls)
        analysis.surface = surface
        analysis.resolution = list(resolution.values())
        analysis.rays = {surface._part._setup[laser_id]: num for laser_id, num in num_rays.items()}
        analysis.name = name
        analysis.fast = False if type == '1' else True
        analysis.id = id
        analysis._added = True
        analysis._urls = []
        analysis._fail_message = None
        analysis._raw_results = {}
        analysis.results = {}
        return analysis

    def _laser_ids_in_rays(self, setup, rays):
        """
        Validates that all laser IDs in the setup's light sources are included in the rays dictionary.

        Args:
        setup: The setup object containing the light sources.
        rays: The dictionary of rays.
        errors: The list to append error messages to.
        """
        # Extract laser IDs from the light sources in the setup
        laser_ids_in_setup = [laser.id for laser in setup.light_sources]

        laser_ids_in_rays = [laser.id for laser in rays.keys()]

        # Check if all laser IDs are in the rays dictionary
        all_laser_ids_in_rays = all(id in laser_ids_in_rays for id in laser_ids_in_setup)

        return all_laser_ids_in_rays

    @property
    def wls(self):
        '''
        Returns a sorted list of the analysis wavelengths.
        '''
        return self._analysis_wls()

    @classmethod
    def _generate_id(cls):
        '''
        Private
        Generates a unique id for the analysis.
        '''
        int_time = int(time.time())
        enc_36_time = np.base_repr(int_time, 36)
        randint = np.base_repr(random.randint(0, 36**5), 36)[2:5]
        id_ = enc_36_time + randint
        return id_

    def _read_file_old(self, file_path):
        '''
        Private.
        Reads the results of the analysis from a file
        '''

        with open(file_path, 'rb') as f:
            content = f.read().strip()


        header = np.frombuffer(content[:v.HEADER_BYTES], dtype=v.ANALYSIS_HEADER_DTYPES)
        data = np.frombuffer(content[v.HEADER_BYTES:], dtype=v.ANALYSIS_MATRIX_DTYPES)
        return header, data

    def _process_v1_file(self, file_path):
        '''
        Deprecated.
        '''
        file_results = {}
        headers_nums, data = self._read_file_old(file_path)
        headers = gu._process_headers_old(headers_nums)
        wls = self.wls

        if v.DEBUG:
            print(f'Headers: {headers}')
            print(f'Wavelengths: {wls}')
            print(f'Data shape: {data.shape}')

        if headers['data_kind'] not in v.SUPPORTED_DATA_KINDS_NAMES:
            return False

        if data.shape[0] != len(wls)*self.resolution[0]*self.resolution[1]:
            return False

        file_results['metadata'] = headers
        res_x = headers['resolution'][0]
        res_y = headers['resolution'][1]
        data_matrices = data.reshape(len(wls), res_x, res_y)

        if v.DEBUG:
            print(f'Data matrices shape: {data_matrices.shape}')

        file_results['data'] = {}
        for i, wl in enumerate(wls):

            if v.DEBUG:
                print(f'Processing wavelength {wl} nm')

            matrix = data_matrices[i]
            file_results['data'][wl] = matrix

        return file_results

    def _process_old(self, directory):
        '''
        Deprecated.
        '''
        file_results = {}
        for file_name in os.listdir(directory):
            if file_name.endswith('.bin'):
                file_path = f'{directory}/{file_name}'
                file_results[file_name] = self._process_v1_file(file_path)
        return file_results

    def _analysis_wls(self):
        '''
        Private.
        Returns a sorted list of the wavelengths of the analysis
        '''
        analysis_wls = []
        setup = self.surface._part._setup
        laser_objects = [setup[laser.id] for laser in self.rays.keys()]
        for laser in laser_objects:
            wls_dicts = laser.data['light_source']['wavelengths_data']
            wls = [wls_dict['wavelength'] for wls_dict in wls_dicts]
            analysis_wls += wls
        analysis_wls = sorted(list(set(analysis_wls)))
        return analysis_wls

    def _extract_file(self,url, destination):
        ''''
        Private.
        Extracts a zip file from a url to a destination folder
        '''

        # Get the zip file from the url
        response = requests.get(url)

        # If the response is successful, extract the zip file
        if response.status_code == 200:

            # Get the zip file as a BytesIO object
            zip_data = BytesIO(response.content)

            # Read the zip file
            with zipfile.ZipFile(zip_data, 'r') as zip_ref:

                # Extract the zip file to the destination folder
                zip_ref.extractall(destination)

        # Return the file path
        file_name = url.split('/')[-1].replace(f'{self.id}_', '').replace('.zip', '')
        file_path = f'{destination}/{file_name}'
        return file_path

    def _unpack(self):
        '''
        Private.
        Unpacks the results of the analysis to a folder
        '''

        # Get the setup id
        setup = self.surface._part._setup.id

        # Create the destination folder
        destination_path = f'.analysis-files/{setup}/{self.surface.id}/{self.id}'
        if not os.path.exists(destination_path):
            os.makedirs(destination_path)

        # Extract the files for each one of the urls
        for url in self._urls:
            self._extract_file(url, destination_path)

        # Return the destination folder
        return destination_path


    ### For version 2 of binary file ###
    def _process_v2_file(self, file_path):
        '''
        This function processes the binary result files of the analysis.
        '''

        # Read the file to get the headers, wls and data
        headers_nums, wls, data = self._read_file2(file_path)

        # Process the headers, and start the file results dictionary
        file_results = self._process_headers(headers_nums)
        file_results_with_wls = self.process_wavelength_data(wls = wls, data = data, file_results = file_results, headers_nums = headers_nums)
        return file_results_with_wls


    ### For version 3 of binary file ###
    def _process_v3_file(self, file_path):
        '''
        This function processes the binary result files of the analysis.
        '''

        # Read the file to get the headers, wls and data
        headers_nums, wls, data = self._read_file3(file_path)

        # Process the headers, and start the file results dictionary
        file_results = self._process_headers(headers_nums)
        file_results_with_wls = self.process_wavelength_data(wls = wls, data = data, file_results = file_results, headers_nums = headers_nums)
        return file_results_with_wls


    def process_wavelength_data(self, wls, data, file_results, headers_nums):
        if v.DEBUG:
            print(f'Headers: {file_results}')
            print(f'Data shape: {data.shape}')

        # Check if the data kind is supported
        if file_results['data_kind'] not in v.SUPPORTED_DATA_KINDS_NAMES:
            if v.DEBUG:
                print(f"Data kind not in the supported kinds: {file_results['data_kind']}")
            return False

        # The data shape should be a matrix of size res_x*res_y for each wavelength. For polarization kind raw, each matrix item has 3 numbers.
        data_shape = (len(wls), self.resolution[0], self.resolution[1], 3) if headers_nums['polarization kind'][0] == 4 else (len(wls), self.resolution[0], self.resolution[1])
        num_items = reduce(lambda x, y: x * y, data_shape)

        # Check if the data shape is correct
        if data.shape[0] != num_items:
            if v.DEBUG:
                print(f"Data shape not correct: {data.shape}, Expected: {data_shape}")
            return False

        # Reshape the long list of numbers to the data matrices
        data_matrices = data.reshape(data_shape)

        if v.DEBUG:
            print(f'Data matrices shape: {data_matrices.shape}')

        # Create a list of results. It's redundant but important since it will be added to a pandas dataframe
        file_results_with_wls = []
        for i, wl in enumerate(wls):
            wl_dict = file_results.copy()
            if v.DEBUG:
                print(f'Processing wavelength {wl} nm')

            # Get the matrix for the current wavelength
            matrix = data_matrices[i]

            # Add the matrix to the results
            wl_dict['data'] = matrix

            # Add the wavelength to the results
            wl_dict['wl'] = wl

            file_results_with_wls.append(wl_dict)

        return file_results_with_wls


    def _process(self, directory):
        '''
        Private.
        Processes the results of the analysis
        '''
        results = []

        # For each results file in the directory, process the file and append the results to a list
        for file_name in os.listdir(directory):
            if file_name.endswith('.bin'):
                file_version = self._check_file_version(f'{directory}/{file_name}')
                self.version = file_version

                if(v.DEBUG):
                    print(f'File Version: {file_version}')

                file_path = f'{directory}/{file_name}'
                if(file_version == 2):
                    file_results = self._process_v2_file(file_path)
                else:
                    file_results = self._process_v3_file(file_path)

                # For now, if the file results are false, we skip it rather then dealing with it.
                if not file_results:
                    continue
                results += file_results

        return results


    def _process_results(self):
        '''
        Private.
        Processes the results of the analysis
        '''
        directory = self._unpack()

        if v.DEBUG:
            print(f'Unpacked analysis files to {directory}')
            print(f'Number of files: {len(os.listdir(directory))}')

        version = self._check_file_version(f'{directory}/{os.listdir(directory)[0]}')

        if version == 1:
            self.version = 1
            results = self._process_old(directory)

        elif version == 2 or version == 3:
            results = self._process(directory)

        else:
            raise Exception(f'Version {version} is not supported')

        self._raw_results = results
        if version == 1:
            self.results = gu.reorganize_analysis_results_dict(self._raw_results.values())

        elif version == 2 or version == 3:
            self.results = AnalysisResults(self._raw_results, maps = self._maps_json, analysis_object = self)

        #delete directory and all subdirectories
        shutil.rmtree(directory)
        self.reset()

        #Sort results by source and polarization - for convenience
        try:
            self.results.sort_values(by = ['spot_target', 'polarization'], axis = 0, inplace = True)
            self.results = self.results.reset_index(drop = True)
        except Exception as e:
            if v.DEBUG:
                print(e)

        return self.results

    def reset(self):
        self._urls = []
        self._raw_results = {}
        return

    def _check_file_version(self, file_path):
        version_bytes  = v.AnalysisProcessVariables.VERSION_BYTES
        with open(file_path, 'rb') as f:
            # version bytes is the range of the 4 bytes indicating the int32 version number. The first int is the beginning, the second is the end
            f.seek(version_bytes[0])
            version = struct.unpack('i', f.read(version_bytes[1] - version_bytes[0]))[0]

            if v.DEBUG:
                print(f'Version: {version}')

        return version


    ### Read V2 binary file ###
    def _read_file2(self, file_path):

        with open(file_path, 'rb') as f:
            content = f.read().strip()

        header = np.frombuffer(content[:v.AnalysisFile2.HEADER_BYTES], dtype=v.AnalysisFile2.HEADER_DTYPES)
        num_wls = header['num_wavelengths'][0]
        if v.DEBUG:
            print(f'Number of wavelengths: {num_wls}')
            print(f'Header nums: {header}')

        wls = np.frombuffer(content[v.AnalysisFile2.HEADER_BYTES:v.AnalysisFile2.HEADER_BYTES + num_wls*4], dtype=[(f'{wl}i', v.AnalysisFile2.WLS_DTYPE) for wl in range(num_wls)])[0]
        data = np.frombuffer(content[v.AnalysisFile2.HEADER_BYTES + num_wls*4:], dtype=v.AnalysisFile2.DATA_DTYPE)

        if v.DEBUG:
            print(f"Data shape: {data.shape}")

        return header, wls, data


    ### Read V3 binary file ###
    def _read_file3(self, file_path):

        with open(file_path, 'rb') as f:
            content = f.read().strip()

        header = np.frombuffer(content[:v.AnalysisFile3.HEADER_BYTES], dtype=v.AnalysisFile3.HEADER_DTYPES)
        num_wls = header['num_wavelengths'][0]
        if v.DEBUG:
            print(f'Number of wavelengths: {num_wls}')
            print(f'Header nums: {header}')

        wls = np.frombuffer(content[v.AnalysisFile3.HEADER_BYTES:v.AnalysisFile3.HEADER_BYTES + num_wls*4], dtype=[(f'{wl}i', v.AnalysisFile3.WLS_DTYPE) for wl in range(num_wls)])[0]
        data = np.frombuffer(content[v.AnalysisFile3.HEADER_BYTES + num_wls*4:], dtype=v.AnalysisFile3.DATA_DTYPE)

        if v.DEBUG:
            print(f'Data shape: {data.shape}')
            print(f'Wavelengths: {wls}')

        return header, wls, data


    def _process_headers(self, headers):
        '''
        Private.
        Processes the headers of the analysis
        '''
        data_kind_mapping = v.DATA_KINDS_MAPPING
        polarization_mapping = v.POLARIZATION_MAPPING


        analysis_kind = v.DATA_KINDS_MAPPING[headers['analysis kind'][0]]
        data_kind = headers['data kind'][0]
        polarization_kind = headers['polarization kind'][0]
        num_hits = headers['num_hits'][0]
        num_wavelengths = headers['num_wavelengths'][0]
        resolution_x = headers['resolution_x'][0]
        resolution_y = headers['resolution_y'][0]
        resolution = (resolution_x, resolution_y)
        data_kind_value = data_kind_mapping.get(data_kind)
        polarization_kind_value = polarization_mapping.get(polarization_kind)
        spot_target_kind = headers['spot_target_kind'][0]
        spot_target_index = headers['spot_target_index'][0]
        if(self.version == 3):
            number_of_surfaces = headers['number_of_surfaces'][0]

        headers = {
            "analysis_kind": analysis_kind,
            "data_kind": data_kind_value,
            "polarization_kind": polarization_kind_value,
            "num_hits": num_hits,
            "num_wavelengths": num_wavelengths,
            "resolution": resolution,
            'spot_target_kind': spot_target_kind,
            'spot_target_index': spot_target_index,

        }
        try: ## Will work if self.version == 3, else this field is not in header
            headers['number_of_surfaces'] = number_of_surfaces
        except Exception as e:
            if v.DEBUG:
                print(f"Field number of surfaces doesnt exist, check version. Error: {e}")

        return headers

    def __str__(self):
        json_dict = {
            "name": self.name,
            "surface": self.surface.id,
            "rays": {laser.id: num for laser, num in self.rays.items()},
            "resolution": tuple(self.resolution),
        }
        string = json.dumps(json_dict, indent = 4)
        return string

    def __eq__(self, other):
        '''
        Equal analyses are analyses with the same parameters of rays, name, resolution and surface.
        '''
        if not isinstance(other, Analysis):
            return False

        is_rays_equal = self.rays == other.rays
        is_name_equal = self.name == other.name
        is_resolution_x_equal = self.resolution[0] == other.resolution[0]
        is_resolution_y_equal = self.resolution[1] == other.resolution[1]
        is_surface_equal = self.surface.id == other.surface.id

        if is_rays_equal and is_name_equal and is_surface_equal and is_resolution_x_equal and is_resolution_y_equal:
            return True

        return False

    @property
    def data(self):
        return self.results

    def show(self, polarizations: list = None, wavelengths: list = None, figsize: Tuple[int, int] = (20, 20), upscale: bool = False):
        '''
        Shows a static figure of the analysis results.
        Args:
            figsize (tuple): The size of the figure.
            upscale (bool): If True, smoothes the pixels over, if the analysis resolution is lower than the figure resolution.

        Returns:
            None

        Shows:
            A figure of the analysis results.

        Raises:
            Exception: If the analysis was not run yet.
        '''

        raise Exception('This method is not supported yet')

        #3DOptix color scale
        cmap = LinearSegmentedColormap.from_list('custom', v.COLOR_SCALE)

        if not self.results:
            raise Exception('Analysis was not run yet')

        if not polarizations:
            one_wl_data = list(self.results.values())[0]
            polarizations = list(one_wl_data.keys())
        if not wavelengths:
            wavelengths = list(self.results.keys())

        #Check if the analysis was run
        if not self.results:
            raise Exception('Analysis was not run yet')

        #Get the number of polarizations and wavelengths of the analysis- polarizations are the rows and wavelengths are the columns of the presented figure
        num_polarizations = len(self.results)
        num_wavelengths = len(self.wls)
        fig = plt.figure(constrained_layout=True, figsize=figsize)
        subfigs = fig.subfigures(nrows=num_polarizations, ncols=1)

        if num_polarizations == 1:
            subfigs = [subfigs]

        for polarization, subfig in zip(polarizations, subfigs):

            subfig.suptitle(f'Polarization {polarization}')
            axs = subfig.subplots(nrows=1, ncols=num_wavelengths)

            if num_wavelengths == 1:
                axs = [axs]

            for wavelength, ax in zip(wavelengths, axs):
                data = self.results[wavelength][polarization]
                if upscale:
                    dpi = plt.rcParams['figure.dpi']
                    data = gu.upscale(data, figsize[0]*dpi, figsize[1]*dpi)
                ax.imshow(data, cmap=cmap)

                ax.set_title(f'Wavelength {wavelength} nm')

        #Show the figure
        plt.show()
        return None

    def show_interactive(self, polarizations: list = None, wavelengths: list = None, figsize: Tuple[int, int] = (20, 20), upscale: bool = False):
        '''
        Shows an interactive figure of the analysis results.
        Args:
            polarizations (list): The polarizations to present. If None, all polarizations are presented.
            wavelengths (list): The wavelengths to present. If None, all wavelengths are presented.
            height (int): The height of the each figure.
            width (int): The width of each figure.
            upscale (bool): If True, smoothes the pixels over, if the analysis resolution is lower than the figure resolution.
        Returns:
            None

        Shows:
            An interactive figure of the analysis results.

        Raises:
            Exception: If the analysis was not run yet.
        '''

        raise Exception('This method is not supported yet')

        dpi = plt.rcParams['figure.dpi']
        height = figsize[0]*dpi
        width = figsize[1]*dpi

        if not self.results:
            raise Exception('Analysis was not run yet')

        if not polarizations:
            one_wl_data = list(self.results.values())[0]
            polarizations = list(one_wl_data.keys())
        if not wavelengths:
            wavelengths = list(self.results.keys())

        for polarization in polarizations:
            for wavelength in wavelengths:
                data = self.results[wavelength][polarization]
                if upscale:
                    data = gu.upscale(data, height, width)

                fig = px.imshow(data, title=f'Polarization {polarization} Wavelength {wavelength} nm', color_continuous_scale = v.COLOR_SCALE)
                fig.update_layout(height=height, width=width)
                fig.show()

    def copy(self):
        '''
        Copies the analysis to a different analysis object with a different id.
        Returns:
        tdo.Analysis: The copied analysis.
        '''
        copied = copy.deepcopy(self)
        copied.id = Analysis._generate_id()
        copied._added = False
        return copied

class AnalysisResults(pd.DataFrame):

    def __init__(self, data, maps, analysis_object):
        '''
        Private.
        Past analysis are stored within the setup.
        When the setup is fetched, the past analysis are created using this method.
        '''
        df = gu.process_results(data, maps)
        super().__init__(df)
        self.attrs['analysis_id'] = analysis_object.id
        return None



import numpy as np
import re

DEBUG = False
BASE_BACKUP = 'DEFAULT'
VALID_RESPONSE_CODES = [200, 201, 202, 204]
# Take the version of the package
VERSION = "5.0.18"


#Base API URL
API_URL = "https://api.3doptix.com/v1"

is_new_session = 'True' #0 for origin that is a human user
class Surfaces:
    DETECTOR_FRONT = 'front'
    DETECTOR_BACK = 'back'
    LENS_FRONT = 'front'
    LENS_BACK = 'back'
    ASPHERE_DEFORMATION = 'ASPHERE'

class Analyses:
    DEFAULT_NUM_RAYS = 30

class Endpoints:
    GET_MATERIALS_ENDPOINT = 'material/name/{material_name}'
    GET_SURFACE_DATA = 'setups/{setup_id}/parts/{part_id}/surfaces/{surface_id}'
    DELETE_ANALYSIS = 'setups/{setup_id}/parts/{part_id}/surfaces/{surface_id}/analyses/{analysis_id}'
    GET_OPTICS_DATA = 'optics/{number_id}'
    CHANGE_CS = 'setups/{setup_id}/parts/{part_id}/cs_data'
    SCATTERING = 'setups/{setup_id}/parts/{part_id}/surfaces/{surface_id}/scattering'

class AddParts:
    VALID_PART_ARGUMENTS = ['pose', 'label']
    VALID_OPTICS_ARGUMENTS = VALID_PART_ARGUMENTS
    VALID_DETECTOR_ARGUMENTS = ['size', 'opacity'] + VALID_PART_ARGUMENTS
    VALID_LIGHT_SOURCE_ARGUMENTS = ['wavelengths', 'add_wavelengths', 'power', 'vis_count', 'vis_count_type', 'rays_direction', 'opacity', 'color', 'gaussian_beam', 'point_source', 'plane_wave'] + VALID_PART_ARGUMENTS


#GET endpoints
GET_SETUPS_ENDPOINT = 'setups'
GET_SETUP_ENDPOINT = 'setups/{setup_id}'
GET_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'
MAX_HISTORY_LEN = 5

# DELETE endpoints
DELETE_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'

#PUT endpoints
PUT_BATCH_CHANGES_ENDPOINT = "setups/{setup_id}/batch_changes"
PUT_SIMULATION_ENDPOINT = 'setups/{setup_id}/simulation'
PUT_PART_ENDPOINT = 'setups/{setup_id}/parts/{part_id}'

#POST endpoints
POST_ADD_ANALYSIS_ENDPOINT = "setups/{setup_id}/parts/{part_id}/surfaces/{surface_id}/analyses"
POST_CREATE_OPTICS_ENDPOINT = "optics"
POST_CREATE_SETUP_ENDPOINT = 'setups'
POST_ADD_PART_ENDPOINT = 'setups/{setup_id}/parts'

#Snellius version: must be compatible
SNELLIUS_VERSION = "2.2.23"
GPU_TYPE = "g4"

#Messages and warnings
SET_API_URL_WARNING = "Are you sure you want to change the API URL? This is internal option and should not be used by users."
SERVER_DOWN_MESSAGE = "The server is down. Please try again later."
WELCOME_MESSAGE = f"Welcome to 3DOptix! 🌈 You are now being connected to the 3DOptix API. Let's start!"
GETTING_PARTS_MESSAGE = f"We are retrieving the setup information. This needs to be done only once. Please wait."

# Setups
SETUP_LABELS = ["General",
                "Microscopy",
                "Telescopy" ,
                "Spectroscopy" ,
                "Imaging" ,
                "Non-linear optics" ,
                "Fiber" ,
                "Illumination" ,
                "Light sources" ,
                "Laser Optics" ,
                "Diffractive Optics"
]

INCH = 'in'
MM = 'mm'
SETUP_UNITS = [MM, INCH]
CREATE_SETUP_UNITS_MAP = {
    INCH: 0,
    MM: 1,
}

class PartType:
    DETECTOR = 9
    LIGHT_SOURCE = 8
    OPTICS = 2
    COORDINATE_SYSTEM = 7
    GENERAL = 0,
    PARAXIAL_LENS = 1,
    CAD = 3,
    GROUP = 5,
    BLACKBOX = 6,
    PART_TYPES = [OPTICS, LIGHT_SOURCE, DETECTOR, COORDINATE_SYSTEM]

## Analysis process variables, exported to mitigate duplications
class AnalysisProcessVariables:
    VERSION_BYTES = [16, 20]

    SPOT_TARGET_KINDS = {
            0: 'Source',
            1: 'Group',
            2: 'Total',
            3: 'None',
        }
    WLS_DTYPE = np.float32
    DATA_DTYPE = np.float32


# Analysis for V2 snellius file
class AnalysisFile2:
    VERSION_BYTES = AnalysisProcessVariables.VERSION_BYTES
    HEADER_DTYPES = [
       ('magic number', np.int32), #0:4
        ('analysis kind', np.int32), #4:8
        ('data kind', np.int32), #8:12
        ('polarization kind', np.int32), #12:16
        ('version', np.int32), #16:20
        ('num_hits', np.int64), #20:28
        ('num_wavelengths', np.int32), #28:32
        ('resolution_x', np.int32), #32:36
        ('resolution_y', np.int32), #36:40
        ('spot_target_kind', np.int32), #40:44
        ('spot_target_index', np.int32), #44:48
        ('detector_width', np.float32), #48:52
        ('detector_height', np.float32), #52:56
    ]
    HEADER_BYTES = 56
    SPOT_TARGET_KINDS = AnalysisProcessVariables.SPOT_TARGET_KINDS

    WLS_DTYPE = AnalysisProcessVariables.WLS_DTYPE
    DATA_DTYPE = AnalysisProcessVariables.DATA_DTYPE

# Analysis for V3 snellius file
class AnalysisFile3:
    VERSION_BYTES = AnalysisProcessVariables.VERSION_BYTES
    HEADER_DTYPES = [
        ('magic number', np.int32), #0:4
        ('analysis kind', np.int32), #4:8
        ('data kind', np.int32), #8:12
        ('polarization kind', np.int32), #12:16
        ('version', np.int32), #16:20
        ('num_hits', np.int64), #20:28
        ('num_wavelengths', np.int32), #28:32
        ('resolution_x', np.int32), #32:36
        ('resolution_y', np.int32), #36:40
        ('spot_target_kind', np.int32), #40:44
        ('spot_target_index', np.int32), #44:48
        ('detector_width', np.float32), #48:52
        ('detector_height', np.float32), #52:56
        ('number_of_surfaces', np.int32),#56:60
    ]
    HEADER_BYTES = 60
    SPOT_TARGET_KINDS = AnalysisProcessVariables.SPOT_TARGET_KINDS

    WLS_DTYPE = AnalysisProcessVariables.WLS_DTYPE
    DATA_DTYPE = AnalysisProcessVariables.DATA_DTYPE



## Analysis file decoding format
ANALYSIS_HEADER_DTYPES = np.dtype([
        ('magic number', np.int32), #0:4
        ('analysis kind', np.int32), #4:8
        ('data kind', np.int32), #8:12
        ('polarization kind', np.int32), #12:16
        ('version', np.int32), #16:20
        ('num_hits', np.int64), #20:28
        ('num_wavelengths', np.int32), #28:32
        ('resolution_x', np.int32), #32:36
        ('resolution_y', np.int32), #36:40
    ])
# ANALYSIS_MATRIX_DTYPES = np.dtype(np.float32).newbyteorder('<')
ANALYSIS_MATRIX_DTYPES = np.float32
HEADER_BYTES = 40
DATA_KINDS_MAPPING = {
    -1: 'INVALID_SPOT_DATA_KIND',
    0: 'SPOT_INCOHERENT_IRRADIANCE_S',
    1: 'SPOT_INCOHERENT_IRRADIANCE_P',
    2: 'SPOT_COHERENT_IRRADIANCE_S',
    3: 'SPOT_COHERENT_IRRADIANCE_P',
    4: 'SPOT_COHERENT_PHASE_S',
    5: 'SPOT_COHERENT_PHASE_P',
    6: 'SPOT_INCOHERENT_IRRADIANCE',
    7: 'SPOT_COHERENT_IRRADIANCE',
    8: 'SPOT_COHERENT_PHASE',
    9: 'SPOT_FIELD',
    10: 'SPOT_MTF',
    11: 'SPOT_WAVEFRONT'
}

SUPPORTED_DATA_KINDS_NAMES = [
    'SPOT_INCOHERENT_IRRADIANCE_S',
    'SPOT_INCOHERENT_IRRADIANCE_P',
    'SPOT_COHERENT_IRRADIANCE_S',
    'SPOT_COHERENT_IRRADIANCE_P',
    'SPOT_COHERENT_PHASE_S',
    'SPOT_COHERENT_PHASE_P',
    'SPOT_INCOHERENT_IRRADIANCE',
    'SPOT_COHERENT_IRRADIANCE',
    'SPOT_COHERENT_PHASE',
]

POLARIZATION_MAPPING = {
    0: 'NONE_POLARIZATION',
    1: 'X_POLARIZATION',
    2: 'Y_POLARIZATION',
    3: 'Z_POLARIZATION',
    4: 'RAW_POLARIZATION',
}

## Valid names
FAST_ANALYSIS_NAMES = [
    "Spot (Incoherent Irradiance)",
    "Spot (Incoherent Irradiance) Through Focus",
    "Grid Distortion",
    "Distortion & Field Curvature",
    "OPD (Optical Path Difference)",
    "Ray Abberations (TRA and LRA)",
    "Polarization Map"
]

DEPRICATED_ANALYSIS_NAMES =  [

    "Spot (Coherent Irradiance) Polarized",
    "Coherent Phase Polarized",
    "Spot (Incoherent Irradiance) Polarized"
]
VALID_ANALYSIS_NAMES =  [
    "Spot (Incoherent Irradiance)",
    "Spot (Coherent Irradiance)",
    "Coherent Phase",
    "Spot (Coherent Irradiance) Huygens",
    "Coherent Phase Huygens",
    "Spot (Coherent Irradiance) Fresnel",
    "Coherent Phase Fresnel",
    "Spot (Coherent Irradiance) Polarized",
    "Coherent Phase Polarized",
    "Spot (Incoherent Irradiance) Polarized"
]
VALID_RESPONSE_CODESANALYSIS_NAMES =  [
    "Spot (Incoherent Irradiance)",
    "Spot (Coherent Irradiance)",
    "Coherent Phase",
    "Spot (Coherent Irradiance) Huygens",
    "Coherent Phase Huygens",
    "Spot (Coherent Irradiance) Fresnel",
    "Coherent Phase Fresnel"
]

## results color scale (as in GUI)
COLOR_SCALE = [
    [0.0, '#0000FF'],
    [0.25, '#00FF00'],
    [0.5, 'yellow'],
    [0.75, 'orange'],
    [1.0, '#FF0000']
]

# Errors
## simulation errors
SIMULATION_ERROR = """Simulation failed (server side).\n Error message: "{message}"."""

## Analyses errors
ANALYSES_ADD_ERROR = "Analyses {not_added} were not added. Please add them first"
ANALYSES_NOT_SAME_SETUP_ERROR = f"Analyses must be from the same setup."
ANALYSIS_RUN_ERROR = """Analyses failed (server side).\n Error message: "{message}"."""
ANALYSES_DUPLICATED_ERROR = """Analyses with ids {duplicated} are duplicated.\nYou can use the existing analyses with the same parameters or force this action by setting 'force' argument to True"""
ANALYSES_RUN_ERROR = """Analysis failed (server side).\n Error message: "{message}"."""

#Arguments
def argument_repair_message(errors):
    return "Invalid arguments. Errors:\n" + "\n".join(errors)

# Analysis
ANALYSIS_RES_RANGE = [0, 10000]
ANALYSIS_RAYS_RANGE = [1, 1e9]

# Detector Data
## size
DETECTOR_SIZE_RANGE = [0.0001, 200]
DETECTOR_OPACITY_RANGE = [0, 1]
#Light source data
##Beam kinds
GAUSSIAN_BEAM = "GAUSSIAN_BEAM"
POINT_SOURCE = "POINT_SOURCE"
PLANE_WAVE = "PLANE_WAVE"
BEAM_KINDS = [POINT_SOURCE, GAUSSIAN_BEAM, PLANE_WAVE]

## Density patterns
XY_GRID = "XY_GRID"
CONCENTRIC_CIRCLES = "CONCENTRIC_CIRCLES"
RANDOM = "RANDOM"
DENSITY_PATTERNS = [XY_GRID, CONCENTRIC_CIRCLES, RANDOM]

## Count type
TOTAL = "TOTAL"
PER_WAVELENGTH = "PER_WAVELENGTH"
LOWER_VIS_COUNT_LIMIT = 1
UPPER_VIS_COUNT_LIMIT = 200
COUNT_TYPES = [TOTAL, PER_WAVELENGTH]

## Plane wave
RECTANGULAR = "RECTANGULAR"
CIRCULAR = "CIRCULAR"
ELLIPTICAL = "ELLIPTICAL"
PLANE_WAVE_TYPES = [RECTANGULAR, CIRCULAR, ELLIPTICAL]

## Appearance
HEX_COLOR_RE = r'#?([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})'
LIGHT_SOURCE_OPACITY_RANGE = [0, 1]
MODEL_RADIUS_RANGE = [1, 10]
## Angular Information
AZIMUTH_Z_RANGE = [-90, 90]
THETA_RANGE = [-180, 180]
PHI_RANGE = [0,180]

## Power
POWER_RANGE = [0, 1e6]

## Point Source
HALF_CONE_ANGLE = "HALF_CONE_ANGLE"
HALF_WIDTH_RECT = "HALF_WIDTH_RECT"
HALF_WIDTH_AT_Z = "HALF_WIDTH_AT_Z"
POINT_SOURCE_TYPES = [HALF_CONE_ANGLE, HALF_WIDTH_RECT, HALF_WIDTH_AT_Z]
PLANE_WAVE_WIDTH_RANGE = [0.0001, 1e5]
PLANE_WAVE_HEIGHT_RANGE = [0.0001, 1e5]
PLANE_WAVE_RADIUS_RANGE = [0.0001, 1e2]
POINT_SOURCE_ANGLE_RANGE = [0.001, 180]
POINT_SOURCE_DIST_Z_RANGE = [0.0001, 9007199254740991]
POINT_SOURCE_DIST_Z_AT_RANGE = [0.0001, 9007199254740991]


## Wavelegnth
WAVELENGTH_RANGE = [140, 20000]
WEIGHTS_RANGE = [1e-5, 1]

# Create Optics
METALLIC_MIRROR = "Metallic Mirror"
DIELECTRIC_MIRROR = "Dielectric Mirror"
DICHROIC_MIRROR = "Dichroic Mirror"
CURVED_MIRROR_CONCAVE = "Curved Mirror (Concave)"
CURVED_MIRROR_CONVEX = "Curved Mirror (Convex)"
CYLINDRICAL_MIRROR_PLANO_CONVEX = "Cylindrical Mirror (Plano-Convex)"
CYLINDRICAL_MIRROR_PLANO_CONCAVE = "Cylindrical Mirror (Plano-Concave)"
PARABOLIC_MIRROR_CONCAVE = "Parabolic Mirror (Concave)"
OFF_AXIS_PARABOLIC_MIRROR = "Off-Axis Parabolic Mirror"
D_SHAPED_MIRROR = "D-Shaped Mirror"
KNIFE_EDGE_PRISM_MIRROR = "Knife-Edge Prism Mirror"
RIGHT_ANGLE_PRISM_MIRROR = "Right-Angle Prism Mirror"
ROOF_PRISM_MIRROR = "Roof Prism Mirror"
ELLIPSOIDAL = "Ellipsoidal"
KNIFE_EDGE_PRISM = "Knife-Edge Prism"
RIGHT_ANGLE_PRISM = "Right-Angle Prism"
EQUIANGULAR_PRISM = "Equilateral Prism"
DOVE_PRISM = "Dove Prism"
WEDGE_PRISM = "Wedge Prism"
PELLIN_BROCA_PRISM = "Pellin Broca Prism"
PENTA_PRISM = "Penta Prism"
AXICON_PRISM = "Axicon Prism"
RHOMBIC_PRISM = "Rhombic Prism"
MULTIPLETS = "Multiplets"
SPHERICAL_LENS_PLANO_CONVEX = "Spherical Lens (Plano-Convex)"
SPHERICAL_LENS_PLANO_CONCAVE = "Spherical Lens (Plano-Concave)"
SPHERICAL_LENS_CONVEX_CONVEX = "Spherical Lens (Convex-Convex)"
SPHERICAL_LENS_CONCAVE_CONCAVE = "Spherical Lens (Concave-Concave)"
SPHERICAL_LENS_CONCAVE_CONVEX = "Spherical Lens (Concave-Convex)"
SPHERICAL_LENS_CONVEX_CONCAVE = "Spherical Lens (Convex-Concave)"
CYLINDRICAL_LENS_PLANO_CONVEX = "Cylindrical Lens (Plano-Convex)"
CYLINDRICAL_LENS_PLANO_CONCAVE = "Cylindrical Lens (Plano-Concave)"
BALL_LENS = "Ball Lens"
HALF_BALL_LENS = "Half Ball Lens"
OFF_AXIS_PARABOLIC_LENS = "Off-Axis Parabolic Lens"
ASPHERIC_LENS = "Aspheric Lens"
ODD_ASPHERE = "Odd Asphere"
PARAXIAL_LENS = "Paraxial lens"
GENERAL_CONIC = "General Conic"
GENERAL_BICONIC = "General Biconic"
LONGPASS_FILTER = "Longpass Filter"
SHORTPASS_FILTER = "Shortpass Filter"
BANDPASS_FILTER = "Bandpass Filter"
DICHROIC_FILTER = "Dichroic Filter"
NOTCH_FILTER = "Notch Filter"
CALIBRATION_FILTER = "Calibration Filter"
SPECTRAL_SHAPE_FILTER = "Spectral Shape Filter"
LASER_LINE_FILTER = "Laser-line Filter"
COLOR_GLASS = "Color Glass"
NEUTRAL_DENSITY_FILTER = "Neutral Density Filter"
HOT_COLD_MIRROR = "Hot/Cold Mirror"
THIN_BS = "Thin BS"
CUBE_BS = "Cube BS"
CUBE_BS_POLARIZED = "Cube BS Polarizing"

THIN_LENS_TYPE = 0,
D_SHAPED_TYPE = 1,
ODD_ASPHERE_TYPE = 2,
CONIC_TYPE = 3,
SPHERICAL_TYPE = 4,
CYLINDRICAL_TYPE = 5,
ELLIPSOIDAL_TYPE = 6,
PARABOLIC_TYPE = 7,
ASPHERIC_TYPE = 8,
HALF_BALL_LENS_TYPE = 9,
BALL_LENS_TYPE = 10,
OFF_AXIS_PARABOLIC_TYPE = 11,
ROOF_PRISM_MIRROR_TYPE = 12,
BS_CUBE_TYPE = 13,
RIGHT_ANGLE_PRISM_TYPE = 14,
EQUAL_PRISM_TYPE = 15,
WEDGE_TYPE = 16,
PENTA_PRISM_TYPE = 17,
DISPERSION_PRISM_TYPE = 18,
PELLIN_BROCA_PRISM_TYPE = 19,
DOVE_PRISM_TYPE = 20,
AXICON_PRISM_TYPE = 21,
RHOMBIC_PRISM_TYPE = 22,
XY_POLYNOMIAL_TYPE = 23,
ODD_POLYNOMIAL_TYPE = 24,
QCON_ASPHERE_TYPE = 25,
Y_TOROID_TYPE = 26,
X_TOROID_TYPE = 27,
GRATING_TYPE = 28,
MULTIPLET_TYPE = 29,
PARAXIAL_LENS_TYPE = 30,
ZERNIKE_TYPE = 31,
MASK_TYPE = 32,
APERTURE_TYPE = 33

class Setups:
    SETUP_LABELS = ["General",
                    "Microscopy",
                    "Telescopy" ,
                    "Spectroscopy" ,
                    "Imaging" ,
                    "Non-linear optics" ,
                    "Fiber" ,
                    "Illumination" ,
                    "Light sources" ,
                    "Laser Optics" ,
                    "Diffractive Optics"
    ]

class Materials:
    SELLEMEIR = "Sellmeier"
    MODIFIED_SELLEMEIR = "Modified Sellmeier"
    CAUCHY = "Cauchy"
    SCHOTT = "Schott"
    CONRADY1 = "Conrady1"
    CONSTANT = "Constant"
    EQUATION_TYPE_MAP = {
        0: CONSTANT,
        1: SELLEMEIR,
        2: MODIFIED_SELLEMEIR,
        3: CAUCHY,
        6: SCHOTT,
        8: CONRADY1,
    }

class eOpticsTypeNames:
    MIRROR = "Mirror"
    LENS = "Lens"
    FILTER = "Filter"
    BEAM_SPLITTER = "Beam Splitter"
    PRISM = "Prism"
    GRATING = "Grating"
    WINDOW = "Window"
    NON_LINEAR_OPTICS = "Nonlinear Optics (Coming soon)"
    OBJECTIVE = "Objective (Coming soon)"
    POLARIZER = "Polarizer (Coming soon)"
    WAVE_PLATE = "Wave Plate (Coming soon)"
    APERTURE = "Aperture"
    MASK = "Mask"
    POLARIZING_ELEMENT = "Polarizing Element"

class eOpticsSubtypes:
    Metallic_Mirror = "Metallic Mirror"
    Dielectric_Mirror = "Dielectric Mirror"
    Dichroic_Mirror = "Dichroic Mirror"
    Curved_Mirror_Concave = "Curved Mirror (Concave)"
    Curved_Mirror_Convex = "Curved Mirror (Convex)"
    Cylindrical_Mirror_Plano_Convex = "Cylindrical Mirror (Plano-Convex)"
    Cylindrical_Mirror_Plano_Concave = "Cylindrical Mirror (Plano-Concave)"
    Parabolic_Mirror_Concave = "Parabolic Mirror (Concave)"
    Off_Axis_Parabolic_Mirror = "Off-Axis Parabolic Mirror"
    D_Shaped_Mirror = "D-Shaped Mirror"
    Knife_Edge_Prism_Mirror = "Knife-Edge Prism Mirror"
    Right_Angle_Prism_Mirror = "Right-Angle Prism Mirror"
    Roof_Prism_Mirror = "Roof Prism Mirror"
    Ellipsoidal = "Ellipsoidal"
    Knife_Edge_Prism = "Knife-Edge Prism"
    Right_Angle_Prism = "Right-Angle Prism"
    Equilateral_Prism = "Equilateral Prism"
    Dove_Prism = "Dove Prism"
    Wedge_Prism = "Wedge Prism"
    Pellin_Broca_Prism = "Pellin Broca Prism"
    Penta_Prism = "Penta Prism"
    Axicon_Prism = "Axicon Prism"
    Rhombic_Prism = "Rhombic Prism"
    Multiplets = "Multiplets"
    Spherical_Lens_Plano_Convex = "Spherical Lens (Plano-Convex)"
    Spherical_Lens_Plano_Concave = "Spherical Lens (Plano-Concave)"
    Spherical_Lens_Convex_Convex = "Spherical Lens (Convex-Convex)"
    Spherical_Lens_Concave_Concave = "Spherical Lens (Concave-Concave)"
    Spherical_Lens_Concave_Convex = "Spherical Lens (Concave-Convex)"
    Spherical_Lens_Convex_Concave = "Spherical Lens (Convex-Concave)"
    Cylindrical_Lens_Plano_Convex = "Cylindrical Lens (Plano-Convex)"
    Cylindrical_Lens_Plano_Concave = "Cylindrical Lens (Plano-Concave)"
    Ball_Lens = "Ball Lens"
    Half_Ball_Lens = "Half Ball Lens"
    Off_Axis_Parabolic_Lens = "Off-Axis Parabolic Lens"
    Aspheric_Lens = "Aspheric Lens"
    Odd_Asphere = "Odd Asphere"
    Paraxial_Lens = "Paraxial lens"
    General_Conic = "General Conic"
    General_Biconic = "General Biconic"
    Longpass_Filter = "Longpass Filter"
    Shortpass_Filter = "Shortpass Filter"
    Bandpass_Filter = "Bandpass Filter"
    Dichroic_Filter = "Dichroic Filter"
    Notch_Filter = "Notch Filter"
    Calibration_Filter = "Calibration Filter"
    Spectral_Shape_Filter = "Spectral Shape Filter"
    Laser_Line_Filter = "Laser-line Filter"
    Color_Glass = "Color Glass"
    Neutral_Density_Filter = "Neutral Density Filter"
    Hot_Cold_Mirror = "Hot/Cold Mirror"
    Thin_BS = "Thin BS"
    Cube_BS = "Cube BS"
    Cube_BS_POLARIZED = "Cube BS Polarizing"
    Blazed_Ruled_Reflective_Grating = "Blazed Ruled Reflective Grating"
    Reflective_Grating = "Reflective Grating"
    Echelle_Grating = "Echelle Grating"
    Transmission_Grating = "Transmission Grating"
    Flat_Window = "Flat Window"
    Wedge_Window = "Wedge Window"
    Curved_Window_Plano_Convex = "Curved Window (Plano-Convex)"
    Curved_Window_Plano_Concave = "Curved Window (Plano-Concave)"
    Elliptical_Aperture = "Elliptical Aperture"
    Rectangular_Aperture = "Rectangular Aperture"
    Aperture_Image_Upload = "Aperture Image Upload"
    Ideal_Polarizer = "Ideal Polarizer"
    Ideal_Waveplate = "Ideal Waveplate"
    GRATING_SUBTYPES = [Blazed_Ruled_Reflective_Grating, Reflective_Grating, Echelle_Grating, Transmission_Grating]

class eBaseShape:
    CIRCULAR = 0
    RECTANGULAR = 1
    VOLUME = 2
    ELLIPTICAL = 3
    BASE_SHAPE_MAP = {0: 'Circular', 1: 'Rectangular', 2: 'Volume', 3: 'Elliptical'}
    GRATING_SHAPES = {'cir': CIRCULAR, 'rec': RECTANGULAR}

class eOpticShape:
    THIN_LENS = 0
    D_SHAPED = 1
    ODD_ASPHERE = 2
    CONIC = 3
    SPHERICAL = 4
    CYLINDRICAL = 5
    ELLIPSOIDAL = 6
    PARABOLIC = 7
    ASPHERIC = 8
    HALF_BALL_LENS = 9
    BALL_LENS = 10
    OFF_AXIS_PARABOLIC = 11
    ROOF_PRISM_MIRROR = 12
    BS_CUBE = 13
    RIGHT_ANGLE_PRISM = 14
    EQUAL_PRISM = 15
    WEDGE = 16
    PENTA_PRISM = 17
    DISPERSION_PRISM = 18
    PELLIN_BROCA_PRISM = 19
    DOVE_PRISM = 20
    AXICON_PRISM = 21
    RHOMBIC_PRISM = 22
    XY_POLYNOMIAL = 23
    ODD_POLYNOMIAL = 24
    QCON_ASPHERE = 25
    Y_TOROID = 26
    X_TOROID = 27
    GRATING = 28
    MULTIPLET = 29
    PARAXIAL_LENS = 30
    ZERNIKE = 31
    MASK = 32
    APERTURE = 33
    OPTICS_SHAPE_MAP = {
        0: 'Thin Lens',
        1: 'D-Shaped',
        2: 'Odd Asphere',
        3: 'General Conic',
        4: 'Spherical',
        5: 'Cylindrical',
        6: 'Ellipsoidal',
        7: 'Parabolic',
        8: 'Aspheric',
        9: 'Half Ball Lens',
        10: 'Ball Lens',
        11: 'Off-Axis Parabolic',
        12: 'Roof Prism Mirror',
        13: 'Cube BS',
        14: 'Right Angle Prism',
        15: 'Equilateral Prism',
        16: 'Wedge Prism',
        17: 'Penta Prism',
        18: 'Dispersion Prism',
        19: 'Pellin Broca Prism',
        20: 'Dove Prism',
        21: 'Axicon Prism',
        22: 'Rhombic Prism',
        23: 'XY Polynomial',
        24: 'Odd Polynomial',
        25: 'Qcon Asphere',
        26: 'Y Toroid',
        27: 'X Toroid',
        28: 'Grating',
        29: 'Multiplet',
        30: 'Paraxial Lens',
        31: 'Zernike',
        32: 'Mask',
        33: 'Aperture'
    }

class APIMessages:
    PART_MODIFIED = 'Part Modified'

class CoordinateSystems:
    WORLD = 0
    WORLD_DATA = {'ref_pose': {'position': [0, 0, 0], 'rotation': [0, 0, 0]}, 'name': 'Global'}

class ScatteringModels:
    LAMBERTIAN = "LAMBERTIAN"
    GAUSSIAN = "GAUSSIAN"
    COS_NTH = "COS_NTH"
    ABG = "ABG"
class Scattering:
    DEFAULT_TRANS = 1
    DEFAULT_ABS = 0
    DEFAULT_REF = 0
    DEFAULT_RELATIVE_POWER = 0.00001
    DEFAULT_NUMBER_OF = 10
    DEFAULT_SIGMAX = 10
    DEFAULT_SIGMAY = 10
    DEFAULT_AZIMUTH_THETA = 0
    DEFAULT_N = 1.5
    DEFAULT_A = 1
    DEFAULT_B = 2
    DEFAULT_G = 2

class Coating:
    TotalAbsorber = 'absorbing'
    IdealReflector = 'ideal_reflector'
    TR_50_50 = 'np_beam_splitter_50_50'
    TR_70_30 = 'np_beam_splitter_70_30'
    TR_30_70 = 'np_beam_splitter_30_70'
    IdealBBAR = 'ideal_broadband_ar'
    COATINGS_TYPES = [TotalAbsorber, IdealReflector, TR_50_50, TR_70_30, TR_30_70, IdealBBAR]



import threading
import pickle
import math
import requests
import pandas as pd
import numpy as np
import random
import time
import requests
import zipfile
import os
import copy
import json

from io import BytesIO
from typing import Any, Union, List, Tuple, Dict, Optional

import threed_optix.package_utils.api as au
import threed_optix.package_utils.vars as v
import threed_optix.package_utils.math as mu
import threed_optix.package_utils.general as gu
import threed_optix.analyses as tdo_analyses
import threed_optix.simulations as tdo_simulations
import threed_optix.parts as tdo_parts
import threed_optix.package_utils.vars as v

class Material:

    @classmethod
    def _new(cls, _api, _data):
        material = cls.__new__(cls)
        material._api = _api
        material._data = _data
        return material

    def __init__(self):
        raise Exception("Material class cannot be instantiated.")

    @property
    def name(self):
        return self._data['name']

    @property
    def id(self):
        return self._data['number_id']

    @property
    def parameters(self):
        return self._data['parameters']

    @property
    def equation(self):
        return v.Materials.EQUATION_TYPE_MAP[self.parameters['type']]

    @property
    def company(self):
        return self.parameters.get('company')

    @property
    def coeffs(self):
        return self.parameters.get('coeffs')

    @property
    def end_lambda(self):
        return self.parameters.get('endLambda')

    @property
    def start_lambda(self):
        return self.parameters.get('startLambda')

    def refractive_index(self, wavelength):

        if 1000*wavelength < self.start_lambda or 1000*wavelength > self.end_lambda:
            raise ValueError(f"Wavelength must be between {self.start_lambda} and {self.end_lambda}")

        if self.equation == v.Materials.SELLEMEIR:
            return Material.calculate_sellmeier(self.coeffs, wavelength)

        if self.equation == v.Materials.CAUCHY:
            return Material.calculate_cauchy(self.coeffs, wavelength)

        if self.equation == v.Materials.CONRADY1:
            return Material.calculate_conrady1(self.coeffs, wavelength)

        if self.equation == v.Materials.CONSTANT:
            return self.coeffs

        if self.equation == v.Materials.MODIFIED_SELLEMEIR:
            return Material.calculate_modified_sellmeier(self.coeffs, wavelength)

        if self.equation == v.Materials.SCHOTT:
            return Material.calculate_schott(self.coeffs, wavelength)

    @staticmethod
    def calculate_sellmeier(coeffs, wavelength):
        wl2 = math.pow(wavelength, 2)
        a = (coeffs['b1'] * wl2) / (wl2 - coeffs['c1'])
        b = (coeffs['b2'] * wl2) / (wl2 - coeffs['c2'])
        c = (coeffs['b3'] * wl2) / (wl2 - coeffs['c3'])
        d = (coeffs['b4'] * wl2) / (wl2 - coeffs['c4'])
        e = (coeffs['b5'] * wl2) / (wl2 - coeffs['c5'])
        f = (coeffs['b6'] * wl2) / (wl2 - coeffs['c6'])
        g = (coeffs['b7'] * wl2) / (wl2 - coeffs['c7'])
        right_side = 1 + a + b + c + d + e + f + g
        refractive_index = math.sqrt(right_side)
        return refractive_index

    @staticmethod
    def calculate_schott(coeffs, wavelength):
        under_sqrt = coeffs['a']\
            + coeffs['b'] * (wavelength ** 2)\
                + coeffs['c'] / (wavelength ** 2)\
                    + coeffs['d'] / (wavelength ** 4)\
                        + coeffs['e'] / (wavelength ** 6)\
                            + coeffs['f'] / (wavelength ** 8)
        return math.sqrt(under_sqrt)

    @staticmethod
    def calculate_cauchy(coeffs, wavelength):
        return Material.calculate_schott(coeffs, wavelength)

    @staticmethod
    def calculate_conrady1(coeffs, wavelength):
        return coeffs['a'] + (coeffs['b']/wavelength) + (coeffs['c']/(wavelength**3.5))

    @staticmethod
    def calculate_modified_sellmeier(coeffs, wavelength):
        under_sqrt = coeffs['a']
        wl2 = wavelength ** 2
        for i in range(1, 8):
            b = coeffs[f'b{i}']
            c = coeffs[f'c{i}']
            under_sqrt += b * wl2 / (wl2 - c)
        return math.sqrt(under_sqrt)

    def describe(self):
        '''
        This function prints the key properties of the material.
        '''
        print(json.dumps({
            'name': self.name,
            'id': self.id,
            'equation': self.equation,
            'company': self.company,
            'end_lambda': self.end_lambda,
            'start_lambda': self.start_lambda
        }, indent=4))

class ThreedOptixAPI:
    """
    Used to manage the communication with the server of 3DOptix.

    Args:
        api_key (str): The API key used for authentication.

    Properties:
        api_key (str): The API key of the user.
        setups (list): The list of setups of the user.
    """


    def __init__(self,
                 api_key: str,
                 verbose: bool = True
                 ):

        # Store the API key
        self.init_error = None
        self.api_key = api_key
        self.setups = None
        self._questions_history = []

        # legacy- not relevat
        self.jobs = []

        # Check if the server is up and the key is valid
        assert self._is_up(), v.SERVER_DOWN_MESSAGE
        assert self._is_key_valid(), v.INVALID_KEY_MESSAGE

        if verbose:
            # Print welcome message
            welcome_thread = threading.Thread(target=au._welcome)
            welcome_thread.start()

        # Fetch setups from the server
        setups_thread = threading.Thread(target=self._initialize_setups)
        setups_thread.start()

        # Wait for both threads to finish
        if verbose:
            welcome_thread.join()

        setups_thread.join()
        if self.init_error is not None:
            #If setups is None, it means that the API key is invalid
            raise self.init_error

    def search_materials(self, material_name):
        # Fetch the material data from the server
        materials_data = au._get_materials(api_key=self.api_key,
                                           material_name=material_name)
        # Create the Material objects
        materials = [Material._new(_api=self, _data=material_data) for material_data in materials_data['materials']]

        return materials

    def create_setup(self,
                     name: str,
                     description: str,
                     labels: List[str],
                     units: str = 'mm',
                     private: bool = True,
                     associated: bool = False,
                     link: str = None,
                     title: str= None,
                     authors: str = None,
                     journal: str = None,
                     abstract: str = None,
                     ):
        '''
        This function creates a new setup with the specified parameters.
        Args:
            name (str): The name of the setup.
            description (str): The description of the setup.
            labels (list): The labels of the setup. enum is in `tdo.SETUP_LABELS`
            units (str): The units of the setup. Default is 'mm'. can be 'mm' or 'inch'
            private (bool): If True, the setup is private. Default is False.
            associated (bool): If True, the setup is associated with a publication. Default is False.
            link (str): The link to the publication. Default is None.
            title (str): The title of the publication. Default is None.
            authors (str): The authors of the publication. Default is None.
            journal (str): The journal of the publication. Default is None.
            abstract (str): The abstract of the publication. Default is None.
        '''


        def verify():
            errors = []

            # Check if the labels are valid
            if not set(labels).issubset(set(v.SETUP_LABELS)):
                errors.append(f"""All labels must one of {', '.join(v.SETUP_LABELS)}""")

            # Check if the units are valid
            if units not in v.SETUP_UNITS:
                errors.append(f"""Units must be one of {', '.join(v.SETUP_UNITS)}""")

            if  not isinstance(name,str) or len(name) == 0:
                errors.append("Name must be a valid non-empty string")

            if not isinstance(description,str) or len(description) == 0:
                errors.append("Description must be a valid non-empty string")

            if len(labels) == 0:
                errors.append("Labels cannot be empty")

            return errors

        # Verify the parameters
        errors = verify()
        # Raise an exception if there are errors
        if len(errors) > 0:
            raise Exception('\n'.join(errors))

        # Organize the data for the API
        data = {
            "details": {
                "setupName": name,
                "labels": labels,
                "generalComments": description,
            },
            "unit": v.CREATE_SETUP_UNITS_MAP[units],
            "permission": 0 if private else 1,
        }

        # Add the associated data if needed
        associated = {
            "isAssociated": associated,
            "link": link,
            "title": title,
            "authors": authors,
            "journal": journal,
            "abstract": abstract,
        }

        for key, value in associated.items():
            if value is not None:
                data['details'][key] = value

        # Create the setup with the API in the database
        response = au._create_setup(data, self.api_key)

        # Get the setup id from the response
        setup_id = response['id']

        # Create the Setup object
        setup = tdo_simulations.Setup._new(_api=self, setup_tuple=(setup_id, name))

        # Add the setup to the setups list
        self.setups.append(setup)
        return setup

    def get(self,
            setup_name: str,
            all: bool = False
            ) -> tdo_simulations.Setup:
        """
        Returns the Setup object with the specified name.
        Args:
            setup_name (str): The name of the setup.
            all (bool): If True, returns all setups with the specified name. else, returns the first one

        Returns:
            Setup (tdo.Setup): The Setup object.
        """

        # If all is True, return all setups with the specified name
        if all:
            setups = []
            for setup in self:
                if setup.name == setup_name:
                    setups.append(setup)
            if len(setups) > 0:
                return setups

        # If all is False, return the first setup with the specified name
        else:
            setup = None
            for s in self.setups:
                if s.name == setup_name:
                    setup = s
                    return setup

        # Raise an exception if the setup is not found
        raise Exception(f"Setup with name {setup_name} not found.")

    def get_setups(self) -> list:
        """
        Returns a list of Setup objects that are associated with the user.

        Returns:
            list: A list of Setup objects.
        """
        return self.setups

    def create_aspheric_lens(self,
                          name,
                          material,
                          diameter,
                          thickness,
                          r1_x,
                          r1_y,
                          r2_x,
                          r2_y,
                          k1_x = 0,
                          k1_y = 0,
                          k2_x = 0,
                          k2_y = 0,
                          coeffs_front = [],
                          coeffs_back = [],
                          check_exist = True,
                          coating = {},
                          db_id = None):

        '''
        This function creates a new biconic lens with the specified parameters.
        Args:
            name (str): The name of the lens.
            material (str | Material): The material of the lens.
            diameter (float): The diameter of the lens.
            thickness (float): The thickness of the lens.
            r1_x (float): The radius of curvature in the x direction of the first surface.
            r1_y (float): The radius of curvature in the y direction of the first surface.
            r2_x (float): The radius of curvature in the x direction of the second surface.
            r2_y (float): The radius of curvature in the y direction of the second surface.
            k1_x (float): The conic constant in the x direction of the first surface. Default is 0.
            k1_y (float): The conic constant in the y direction of the first surface. Default is 0.
            k2_x (float): The conic constant in the x direction of the second surface. Default is 0.
            k2_y (float): The conic constant in the y direction of the second surface. Default is 0.
            coeffs_front (list): The coefficients of the front surface. Default is [].
            coeffs_back (list): The coefficients of the back surface. Default is [].
            check_exist (bool): If True, checks if the lens already exists. if it exists, it returns it and doesn't create a new one. Default is True.

        Returns:
            str: The database id of the created lens.
        '''

        type = v.eOpticsTypeNames.LENS
        subtype = v.eOpticsSubtypes.General_Biconic
        shape = v.eOpticShape.CONIC
        base_shape = v.eBaseShape.CIRCULAR

        deformation = {}
        if len(coeffs_front) > 0:
            deformation[v.Surfaces.LENS_FRONT] =  [{
                "type": v.Surfaces.ASPHERE_DEFORMATION,
                "params": {
                    "terms": [[c[0], c[0], c[1]] for c in coeffs_front]
                }
                }
                ]
        if len(coeffs_back) > 0:
            deformation[v.Surfaces.LENS_BACK] = [{
                "type": v.Surfaces.ASPHERE_DEFORMATION,
                "params": {
                    "terms": [[c[0], c[0], c[1]] for c in coeffs_back]
                }
                }
                ]

        # Create the geometry dictionary
        geometry = {
            "diameter": diameter,
            "thickness_center": thickness,
            "r1_x": r1_x,
            "r1_y": r1_y,
            "r2_x": r2_x,
            "r2_y": r2_y,
            "k1_x": k1_x,
            "k1_y": k1_y,
            "k2_x": k2_x,
            "k2_y": k2_y,
            "deformation": deformation
            }

        # Create the lens with the API
        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material
        )


    def create_biconic_lens(self,
                          name,
                          material,
                          diameter,
                          thickness,
                          r1_x,
                          r1_y,
                          r2_x,
                          r2_y,
                          k1_x = 0,
                          k1_y = 0,
                          k2_x = 0,
                          k2_y = 0,
                          check_exist = True,
                          coating = {},
                          db_id = None):

        '''
        This function creates a new biconic lens with the specified parameters.
        Args:
            name (str): The name of the lens.
            material (str | Material): The material of the lens.
            diameter (float): The diameter of the lens.
            thickness (float): The thickness of the lens.
            r1_x (float): The radius of curvature in the x direction of the first surface.
            r1_y (float): The radius of curvature in the y direction of the first surface.
            r2_x (float): The radius of curvature in the x direction of the second surface.
            r2_y (float): The radius of curvature in the y direction of the second surface.
            k1_x (float): The conic constant in the x direction of the first surface. Default is 0.
            k1_y (float): The conic constant in the y direction of the first surface. Default is 0.
            k2_x (float): The conic constant in the x direction of the second surface. Default is 0.
            k2_y (float): The conic constant in the y direction of the second surface. Default is 0.
            check_exist (bool): If True, checks if the lens already exists. if it exists, it returns it and doesn't create a new one. Default is True.

        Returns:
            str: The database id of the created lens.
        '''

        type = v.eOpticsTypeNames.LENS
        subtype = v.eOpticsSubtypes.General_Biconic
        shape = v.eOpticShape.CONIC
        base_shape = v.eBaseShape.CIRCULAR

        # Create the geometry dictionary
        geometry = {
            "diameter": diameter,
            "thickness_center": thickness,
            "r1_x": r1_x,
            "r1_y": r1_y,
            "r2_x": r2_x,
            "r2_y": r2_y,
            "k1_x": k1_x,
            "k1_y": k1_y,
            "k2_x": k2_x,
            "k2_y": k2_y,
            }

        # Create the lens with the API
        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material
        )

    def create_conic_lens(self,
                          name,
                          material,
                          diameter,
                          thickness,
                          r1,
                          r2,
                          k1 = 0,
                          k2 = 0,
                          check_exist = True,
                          coating = {},
                          db_id = None
                          ):
        '''
        This function creates a new conic lens with the specified parameters.
        Args:
            name (str): The name of the lens.
            material (str | Material): The material of the lens.
            diameter (float): The diameter of the lens.
            thickness (float): The thickness of the lens.
            r1 (float): The radius of curvature of the lens.
            r2 (float): The radius of curvature of the lens.
            k1 (float): The conic constant of the lens. Default is 0.
            k2 (float): The conic constant of the lens. Default is 0.
            check_exist (bool): If True, checks if the lens already exists. if it exists, it returns it and doesn't create a new one. Default is True.

        Returns:
            str: The database id of the created lens.
        '''

        type = v.eOpticsTypeNames.LENS
        subtype = v.eOpticsSubtypes.General_Conic
        shape = v.eOpticShape.CONIC
        base_shape = v.eBaseShape.CIRCULAR

        # Create the geometry dictionary
        geometry = {
            "diameter": diameter,
            "thickness_center": thickness,
            "r1": r1,
            "r2": r2,
            "k1": k1,
            "k2": k2
            }

        # Create the lens with the API
        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material
        )

    def create_spherical_lens(self,
                          name,
                          material: Union[str, Material],
                          diameter,
                          thickness,
                          r1,
                          r2 = None,
                          check_exist = True,
                          coating = {},
                          db_id = None):
        '''
        This function creates a new spherical lens with the specified parameters.
        Args:
            name (str): The name of the lens.
            material (str | Material): The material of the lens.
            diameter (float): The diameter of the lens.
            thickness (float): The thickness of the lens.
            r1 (float): The radius of curvature of the lens.
            r2 (float): The radius of curvature of the lens. Default is None.
            check_exist (bool): If True, checks if the lens already exists. if it exists, it returns it and doesn't create a new one. Default is True.

        Returns:
            str: The database id of the created lens.
        '''



        type = v.eOpticsTypeNames.LENS

        ## Check if name is defined
        if(not name or not isinstance(name,str)):
            raise ValueError('Name must be a valid string')

        # Check if the radii are valid
        if r1 == 0 or r2 == 0:
            raise ValueError('radii cannot be 0. For plano surface, use r1 when r2 is None.')

        # Determine the subtype based on the radii
        if r2 is None:
            # if r2 is None, the lens is plano
            if r1 > 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Plano_Convex
            elif r1 < 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Plano_Concave

        elif r1 < 0:
            # if r1 is negative, the lens is concave
            if r2 < 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Concave_Convex
            elif r2 > 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Concave_Concave
        elif r1 > 0:
            # if r1 is positive, the lens is convex
            if r2 < 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Convex_Convex
            elif r2 > 0:
                subtype = v.eOpticsSubtypes.Spherical_Lens_Convex_Concave

        shape = v.eOpticShape.SPHERICAL
        base_shape = v.eBaseShape.CIRCULAR

        # Create the geometry dictionary
        geometry = {
            "diameter": diameter,
            "thickness_center": thickness,
            "r1": r1,
            }

        # Add r2 to the geometry dictionary if it is not None
        if r2 is not None:
            geometry["r2"] = r2

        # Create the lens with the API
        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material
        )

    def create_ball_lens(self,
                          name,
                          material,
                          diameter,
                          half = False,
                          check_exist = True,
                          coating = {},
                          db_id = None):
        '''
        Not implemented yet.
        '''

        type = v.eOpticsTypeNames.LENS
        shape = v.eOpticShape.CONIC
        base_shape = v.eBaseShape.CIRCULAR
        subtype = v.eOpticsSubtypes.Half_Ball_Lens if half else v.eOpticsSubtypes.Ball_Lens
        geometry = {
            "diameter": diameter,
            }

        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material
        )

    def create_grating(self,
                       name,
                       material,
                       shape,
                       grooves,
                       order,
                       subtype,
                       orientation_vector,
                       grating_side = 'Front',
                       check_exist = True,
                       coating = {},
                       db_id = None,
                       **kwargs
                       ):
        '''
        This function creates a new grating with the specified parameters.
        Args:
            name (str): The name of the grating.
            material (str | Material): The material of the grating.
            shape (str): The shape of the grating. can be either 'cir' or 'rect'.
            grooves (int): The number of grooves of the grating.
            order (int): The order of the grating.
            subtype (str): The subtype of the grating. enum is in `tdo.GRATING_SUBTYPES`
            orientation_vector (list): The normalized orientation vector of the grating.
            grating_side (str): The side of the grating. can be "Back" or "Front". Default is 'Front'.
            check_exist (bool): If True, checks if the grating already exists. if it exists, it returns it and doesn't create a new one. Default is True.
            coating (dict): The coating of the grating. can be an empty dictionary or a dictionary with surface names as keys and `tdo.Coating` enum as values. Default is {}.
            db_id (str): The database id of the grating. If defined, it will edit the optics in the database. Default is None.
            kwargs (dict): The additional parameters of the grating. For circular grating, 'diameter' and 'thickness' must be provided. For rectangular grating, 'width', 'height', and 'thickness' must be provided. For blazed ruled reflective grating, 'blaze_angle' and 'blaze_wavelength' must be provided.

        Returns:
            str: The database id of the created grating.
        '''

        def verify():

            if v.DEBUG:
                print(f'Kwargs are {kwargs}')
                print(f'Shape is {shape}')
                print(f'Subtype is {subtype}')

            errors = []

            # Check if the subtype is valid
            if subtype not in v.eOpticsSubtypes.GRATING_SUBTYPES:
                errors.append(f"""Subtype must be one of {', '.join(v.eOpticsSubtypes.GRATING_SUBTYPES)}""")

            # Check if the shape is valid
            if shape not in v.eBaseShape.GRATING_SHAPES.keys():
                errors.append(f"""Shape must be one of {', '.join(v.eBaseShape.GRATING_SHAPES.keys())}""")

            if shape == 'cir':
                # Check if the diameter and thickness are provided
                if any([key not in kwargs for key in ['diameter', 'thickness']]):
                    errors.append("For circular grating, diameter and thickness must be provided")
            elif shape == 'rect':
                # Check if the width, height, and thickness are provided
                if any([key not in kwargs for key in ['width', 'height', 'thickness']]):
                    errors.append("For rectangular grating, width, height and thickness must be provided")

            if subtype in [v.eOpticsSubtypes.Blazed_Ruled_Reflective_Grating,
                       v.eOpticsSubtypes.Echelle_Grating,
                       v.eOpticsSubtypes.Transmission_Grating]:
                # Check if the blaze angle and blaze wavelength are provided
                if any([key not in kwargs for key in ['blaze_angle', 'blaze_wavelength']]):
                    errors.append("Blaze angle and blaze wavelength must be provided for the selected subtype")

            # Check if the orientation vector is in length 3
            if len(orientation_vector) != 3:
                errors.append("Orientation vector must have 3 elements")

            # Check if the orientation vector has a length of 1
            if not mu.is_normalized(orientation_vector):
                errors.append("Orientation vector must be normalized")

            # Check if the grating side is valid
            if grating_side not in ['Front', 'Back']:
                errors.append("Grating side must be one of 'front' or 'back'")

            return errors

        # Verify the parameters
        errors = verify()
        if len(errors) > 0:
            raise Exception('\n'.join(errors))

        # Create the geometry dictionary based on the shape
        if shape == 'cir':
            geometry = {
                "diameter": kwargs['diameter'],
                'thickness': kwargs['thickness'],
            }
        elif shape == 'rec':
            geometry = {
                "width": kwargs['width'],
                'height': kwargs['height'],
                'thickness': kwargs['thickness'],
            }


        type = v.eOpticsTypeNames.GRATING
        base_shape = v.eBaseShape.GRATING_SHAPES[shape]
        shape = v.eOpticShape.GRATING

        if subtype in [v.eOpticsSubtypes.Blazed_Ruled_Reflective_Grating,
                       v.eOpticsSubtypes.Echelle_Grating,
                       v.eOpticsSubtypes.Transmission_Grating]:
            physical_data = {
                "blaze_angle": kwargs['blaze_angle'],
                "blaze_wavelength": kwargs['blaze_wavelength'],
            }
        else:
            physical_data = {}

        physical_data['grooves'] = grooves
        physical_data['order'] = order
        physical_data['orientation_vector'] = {'x': orientation_vector[0], 'y': orientation_vector[1], 'z': orientation_vector[2]}
        physical_data['grating_side'] = grating_side

        # Create the grating with the API
        return self._create_optics(
            type=type,
            subtype=subtype,
            base_shape=base_shape,
            geometry=geometry,
            shape=shape,
            name=name,
            coating=coating,
            check_exist=check_exist,
            db_id=db_id,
            material=material,
            physical_data=physical_data
        )

    def _create_optics(self, type, subtype, base_shape, material, shape, geometry, name, check_exist, db_id, coating, physical_data = None):
        '''
        Private.
        This function is the base function for creating optics.
        It creates the optics with the specified parameters.
        '''

        parameters = {
            "type": type,
            "subType": subtype,
            "materialID": material.id if isinstance(material, Material) else material,
            "baseShape": base_shape,
            "shape": shape,
            "coating": coating,
            "geometry": geometry,
        }

        if physical_data is not None:
            parameters['physical_data'] = physical_data

        optics = {
            "name": name,
            "parameters": parameters
        }

        if db_id is not None:
            optics['number_id'] = db_id

        data = {
            "optics": optics,
            "checlExist": check_exist
        }

        return au._create_part(parameters=data, api_key=self.api_key).get('number_id')

    def ask(self,
            question: str
            ):
        '''
        Not implemented yet.
        This function will call the SDK assistant to answer the question about the SDK
        '''
        try:
            # Update the history with the user question
            self._questions_history.append({"role": "user", "content": question})
            # Ask the question to the assistant
            response = au._ask(self._questions_history[-v.MAX_HISTORY_LEN:], self.api_key)
        except Exception as e:
            # If the request fails, remove the question from the history and raise the exception
            self._questions_history.pop()
            raise e

        # Update the history with the assistant response
        self._questions_history.append({"role": "assistant", "content": response})
        return response

    def feedback(self, feedback: str):
        '''
        Not implemented yet.
        Sending feedback to us about the SDK.
        '''
        raise NotImplementedError("This function is not implemented yet.")

    def _add_part(self, setup_id, type_, db_id = None):
        '''
        Private.
        This is the base function for adding parts to the setup.
        It adds the part with the specified parameters to the setup with the specified id.
        The setup calls this function as a communication bridge.
        '''


        data = {
            "type": type_,
        }
        if db_id is not None and type_ == v.PartType.OPTICS:
            data['number_id'] = db_id

        response = au._add_part(setup_id, data, self.api_key)
        part_id = response['id']
        return part_id

    def _delete_part(self, setup_id, part_id):
        '''
        Private.
        This is the base function for deleting parts from the setup.
        It deletes the part with the specified id from the setup with the specified id.
        The setup calls this function as a communication bridge.
        '''
        return au._delete_part(setup_id, part_id, self.api_key)

    def __contains__(self, item: Union[str, tdo_simulations.Setup]) -> bool:
        """
        Allows checking if a setup id is in the API.

        Args:
            item (tdo.Setup): The setup name, id, or object.

        Returns:
            bool: True if the setup exists, False otherwise.
        """
        contains = False
        if isinstance(item, str):
            for setup in self:
                if setup.id == item:
                    contains = True
                    break
        elif isinstance(item, tdo_simulations.Setup):
            for setup in self:
                if setup.id == item.id:
                    contains = True
                    break
        return contains

    def __len__(self) -> int:
        '''
        Returns the number of setups in the API.
        '''
        return len(self.setups)

    def __iter__(self):
        """
        Iterates over the setups of the API.
        """
        return iter(self.setups)

    def __str__(self) -> str:
        '''
        Prints the API object and the setups it contains.
        '''
        string = f"Client with {len(self)} setups:\n"
        for setup in self:
            string += f"  - {setup.name} ({setup.id})\n"
        return string

    def __getitem__(self, key: str) -> tdo_simulations.Setup:
        """
        Args:
            key (str): The id of the requested setup.
        Returns:
            Setup (tdo.Setup): The requested Setup object.
        """
        # if isinstance(key, int):
        #     return self.setups[key]
        if isinstance(key, str):
            for setup in self:
                if setup.id == key:
                    return setup
            raise KeyError(f"Setup with id {key} not found.")
        raise TypeError(f"Invalid key type {type(key)}. Must be setup id.")

    def _initialize_setups(self):
        '''
        Private.
        Shouldn't be called directly. Initializes the setups property.
        '''
        self.setups = self._get_setups_info()
        if self.setups is not None:
            self.setups = [tdo_simulations.Setup._new(_api=self, setup_tuple=setup) for setup in self.setups]
        return self.setups

    def _get_surface(self, setup_id, surface_id, part_id):
        '''
        Private.
        Shouldn't be called directly. Returns the surface with the specified id.
        '''
        surface_data = au._get_surface(self.api_key, setup_id, surface_id, part_id)
        if surface_data[0] is None:
            raise Exception(f"Surface with id {surface_id} not found.")
        return surface_data[0]

    def _get_setups_info(self) -> list:
        """
        Private.
        Shouldn't be called directly. Returns a list of setups info.
        """
        try:
            data, message = au._get_setups(self.api_key)

            if data is None:
                self.init_error = message

            infos = []
            for info_json in data['setups']:
                infos.append((info_json['id'], info_json['name']))
            return infos
        except Exception as e:
            self.init_error = e
            return None

    def _get_setup_parts(self, setup_id: str) -> list:
        '''
        Private.
        Shouldn't be called directly. Returns a list of parts of the specified setup.
        '''
        parts = au._get_setup(setup_id, self.api_key)
        parts = parts[0]
        return parts

    def _get_part(self, part_id: str, setup_id) -> dict:
        '''
        Private.
        Shouldn't be called directly. Returns a dictionary of the specified part.
        '''
        part = au._get_part(setup_id,part_id,  self.api_key)
        if part[0] is None:
            raise Exception(f"Part with id {part_id} not found.")

        return part[0]

    def _extract_setup_object(self, setup):
        '''
        Private.
        Shouldn't be called directly. Returns the setup object from the specified setup.
        '''
        if isinstance(setup, str):

            if setup in self:
                setup_object = self[setup]
                return setup_object

            elif self.get(setup) is not None:
                setup_object = self.get(setup)
                return setup_object

            else:
                raise Exception(f"Setup with id or name {setup} not found.")

        if isinstance(setup, tdo_simulations.Setup):
            return setup

        if isinstance(setup, int):
            return self.setups[setup]

        raise TypeError(f"Invalid setup type {type(setup)}. Must be Setup object, name, index or id.")

    def _is_up(self) -> bool:
        """
        Calls _healthcheck and returns a boolean indicating if the server is up.

        Returns:
            bool: True if the server is up, False otherwise.
        """
        return au._healthcheck()[0]

    def _is_key_valid(self):
        '''
        Not implemented yet.
        '''
        return True

    def _update_part(self, setup_id: str, part: tdo_parts.Part) -> tuple:
        '''
        Modifies the part's data to the setup with that id.
        Args:
            setup_id (str): The id of the setup.
            part (Part): The part object to update.

        Returns:
            tuple: A tuple of (success, message), where:
                success (bool): True if the part was updated successfully, False otherwise.
                message (str): The message from the server.
        Raises:
            Exception: If the part was not updated successfully.
        '''
        data, message = au._change_part(setup_id, part.id, part._changes, self.api_key)

        if v.DEBUG:
            print(f'Success is: {data}')
            print(f'Message is: {message}')

        if data is None:
            raise Exception(message)
        return (data, message)

    def _run(self,
            setup: tdo_simulations.Setup
            ) -> tdo_analyses.RayTable:
        """
        Propagate the rays in the setup and returns a RayTable object.

        Args:
            setup (tdo.Setup): The setup to run.

        Returns:
            RayTable (tdo.analyses.RayTable): The RayTable object with the results.

        Raises:
            Exception: If the simulation failed.
        """
        setup_object = self._extract_setup_object(setup)

        data, message = au._run_simulation(setup_object.id, self.api_key)

        if data == None:
            raise Exception(message)
        if data['results']['error']['code'] != 0:
            raise Exception(v.SIMULATION_ERROR.format(message = data['results']['error']['message']))

        if data is not None:
            ray_table = data['results']['data']['ray_table']
            maps_url = data['maps_url']
            return tdo_analyses.RayTable(ray_table, maps_url, setup_object)
        else:
            raise Exception(message)

    def _run_async(self, setup: tdo_simulations.Setup):
        '''
        Not implemented yet.
        '''
        setup_object = self._extract_setup_object(setup)
        data, message = au._run_simulation(setup_object.id, self.api_key, is_sync = False)
        if data == None:
            raise Exception(message)
        if data['results']['error']['code'] != 0:
            raise Exception(v.SIMULATION_ERROR.format(message = data['results']['error']['message']))

        if data is not None:
            ray_table = data['results']['data']['ray_table']
            maps_url = data['maps_url']
            return tdo_analyses.RayTable(ray_table, maps_url, setup_object)
        else:
            raise Exception(message)

    def _run_batch(self,
                  setup: tdo_simulations.Setup,
                  configuration: dict
                  ):
        """
        Not implemented yet.
        """
        response = au._run_batch(setup.id, configuration, self.api_key)
        if response[0] is not None:
            json_ = response[0]
            json_['number_of_changes'] = configuration['number_of_changes']
            json_['simulation_file_prefix'] = configuration['simulation_file_prefix']
            job = tdo_simulations.Job._from_json(response[0], _api = self, _setup = setup)
            job._url = job._url.replace('$', '')
            self.jobs.append(job)
            return job
        else:
            raise Exception(response[1])

    def _run_analysis(self,
                     analysis: tdo_analyses.Analysis,
                     auto_add: bool = False,
                     force: bool = False
                     ) -> bool:
        if auto_add:
            self._add_analysis(analysis, force = force)
        data, _ =  au._run_analysis(setup_id=analysis.surface._part._setup.id, api_key=self.api_key, analysis_id=analysis.id)
        if v.is_new_session == 'True':
            if data['results']['error']['code'] != 0:
                raise Exception(data['results']['error']['message'])

            maps_url = data['maps_url']
            analysis._maps_json = requests.get(maps_url).json()

            if v.DEBUG:
                print(f'Maps json is {analysis._maps_json}')


            analysis_datos = data['results']['data']['analysis']
            analysis._urls = [data['url'] for data in analysis_datos]

            if v.DEBUG:
                print(f'Analysis urls is {analysis._urls}')

            analysis.results = analysis._process_results()

            return copy.deepcopy(analysis.results)

        if v.is_new_session == 'False':
            return {'status': 'success', 'message': f'Analysis {analysis.id} ran successfully', 'data': data}


    def _run_analyses(self,
                     analyses: list,
                     auto_add: bool = False,
                     force:bool = False
                     ):
        '''
        Private.
        '''
        ids = [analysis.id for analysis in analyses]
        setup_id = analyses[0].surface._part._setup.id

        response = {'success': [], 'failed': []}

        if not all([analysis.surface._part._setup.id == setup_id for analysis in analyses]):
            raise Exception(v.ANALYSES_NOT_SAME_SETUP_ERROR)

        if not auto_add and force:
            raise Exception("Force argument can only be used with auto_add argument")

        if auto_add:
            added_successfully = []
            added_failed = []

            for analysis in analyses:
                added = self._add_analysis(analysis, force = force)

                if added:
                    added_successfully.append((analysis, analysis.id))

                else:
                    print(f"Analysis {analysis.id} was failed to be added")
                    added_failed.append((analysis, analysis.id))

            response['added'] = added_successfully
            response['failed'] += added_failed
            ids = [analysis[1] for analysis in added_successfully]

        ran_successfully = []
        ran_failed = []

        if not all([analysis._added for analysis in analyses]):
            not_added = [analysis.id if analysis._added else None for analysis in analyses]
            raise Exception(v.ANALYSES_ADD_ERROR.format(not_added = not_added))

        data, message =  au._run_analyses(setup_id, self.api_key, ids)

        if data == None:
            raise Exception(message)

        for i, analysis in enumerate(analyses):
            is_successful = data['results']['error']['code'] == 0
            ran_successfully.append((analysis, analysis.id))
            if is_successful:
                analysis_datos = data['results']['data']['analysis']
                for data in analysis_datos:
                    url = data['url']
                    analysis._urls.append(url)
                analysis.results = analysis._process_results()
            else:
                ran_failed.append((analysis, analysis.id, data['results']['error']['message']))

        if len(ran_failed) > 0:
            messages = '\n'.join([f'Analysis {failed[1]}: {failed[2]}' for failed in ran_failed])
            raise Exception(v.ANALYSES_RUN_ERROR.format(message = messages))
        return analysis.results

    def _add_analysis(self,
                     analysis: tdo_analyses.Analysis,
                     force = False) -> bool:
        '''
        Adds the specified analysis to the setup.
        Args:
            analysis (tdo.Analysis): The analysis to add.
            force (bool): If True, adds the analysis to the setup even if identical analysis already exists.
            Default is False.
            Each analysis stores its result seperately.
            Filling the same analysis twice will result in larger user's memory resources consumption.
        Returns:
            bool: True if the analysis was added successfully, False otherwise.
        Raises:
            Exception:
                - If the server failed to respond.
                - If the analysis is duplicated and force is False.

        '''
        if analysis.id in [a.id for a in analysis.surface.analyses]:
            analysis._added = True
            return None

        if not force:
            is_duplicated = False
            duplicated = []

            for existing_analysis in analysis.surface.analyses:
                if analysis == existing_analysis:
                    is_duplicated = True
                    duplicated.append(analysis.id)
            if is_duplicated:
                raise Exception(v.ANALYSES_DUPLICATED_ERROR.format(duplicated = ", ".join(duplicated)))

        data = {
                "name": analysis.name,
                "num_rays": {laser.id: num for laser, num in analysis.rays.items()},
                "resolution": {"x": analysis.resolution[0], "y": analysis.resolution[1]},
                "type": 0 if analysis.fast == True else 1,
            }
        part = analysis.surface._part
        response = au._add_analyses(setup_id=part._setup.id,
                                    part_id=part.id,
                                    surface_id=analysis.surface.id,
                                    data=data,
                                    api_key=self.api_key)
        if response[1] == 'Analyses successfully added':
            return response[2]
        raise Exception(response[1])

    def _delete_analysis(self, setup_id, part_id, surface_id, analysis_id):
        '''
        Private.
        '''
        return au._delete_analysis(setup_id=setup_id, part_id=part_id, surface_id=surface_id, analysis_id=analysis_id, api_key=self.api_key)

    def optics_data(self, db_id: Union[str, tdo_parts.Part]):
        '''
        This function returns the optics data from the database.
        Args:
            db_id (str | Part): The database id of the optics or the Part object.

        Returns:
            dict: The optics data.
        '''

        # Get the db_id from the Part object if it is provided
        if isinstance(db_id, tdo_parts.Part):
            db_id = db_id.db_id

        r = au._get_optics_data(db_id = db_id, api_key = self.api_key)
        if r[0] is None:
            raise Exception(r[1])
        data = r[0][db_id]

        # Convert the data from category numbers to strings
        data['parameters']['shape'] = v.eOpticShape.OPTICS_SHAPE_MAP[data['parameters']['shape']]
        data['parameters']['baseShape'] = v.eBaseShape.BASE_SHAPE_MAP[data['parameters']['baseShape']]
        return data

    def _change_cs(self, setup_id, part_id, lcs_id, rcs_id):
        '''
        Private.
        This function changes the coordinate system of the part.
        Used by the Part object, this is a communication bridge.
        Args:
            setup_id (str): The id of the setup.
            part_id (str): The id of the part.
            lcs_id (str): The id of the local coordinate system.
            rcs_id (str): The id of the reference coordinate system.

        Returns:
            tuple: A tuple of (success, message), where:
                data (bool): True if the coordinate system was changed successfully, False otherwise.
                message (str): The message from the server.
        '''

        data = {
            "lcs": lcs_id,
        }
        if rcs_id != v.CoordinateSystems.WORLD:
            data['rcs'] = rcs_id

        return au._change_cs_data(setup_id=setup_id, part_id=part_id, data=data, api_key=self.api_key)

    def _change_scattering(self, setup_id, part_id, surface_id, scattering):
        '''
        This function changes the scattering data of the surface.
        Args:
            setup_id (str): The id of the setup.
            part_id (str): The id of the part.
            surface_id (str): The id of the surface.
            scattering (dict): The scattering data.

        Returns:
            tuple: A tuple of (data, message), where:
                data (bool): True if the scattering data was changed successfully, False otherwise.
                message (str): The message from the server.
        '''
        return au._change_scattering(setup_id=setup_id, part_id=part_id, surface_id=surface_id, data=scattering, api_key=self.api_key)

class Client(ThreedOptixAPI):
    '''
    This is the main class of the SDK.
    It is used to interact with the API and create the setups.
    It wraps ThreedOptixAPI from legacy reasons.
    '''
    def __init__(self, api_key: str, verbose: bool = True):
        super().__init__(api_key, verbose)



import numpy as np
from scipy.stats import norm
import cv2
import threed_optix.package_utils.vars as v
# import matlabparser as mpars
# import threed_optix.package_utils.matlab as mt
import re
import matplotlib.pyplot as plt
import plotly.express as px

def wavelengths_normal_distribution(mean_wavelength, std_dev, num_wavelengths):
    wavelengths = {}
    wavelengths_list = np.linspace(mean_wavelength - 3 * std_dev, mean_wavelength + 3 * std_dev, num_wavelengths)
    weights =  norm.pdf(wavelengths_list, mean_wavelength, std_dev)
    sum_of_weights = np.sum(weights)

    for i in range(num_wavelengths):
        wavelength = wavelengths_list[i]
        weight = weights[i] * (v.WEIGHTS_RANGE[1] - v.WEIGHTS_RANGE[0]) / sum_of_weights + v.WEIGHTS_RANGE[0]
        weight = np.maximum(v.WEIGHTS_RANGE[0], np.minimum(v.WEIGHTS_RANGE[1], weight))
        wavelengths[wavelength] = weight

    return wavelengths

def wavelengths_uniform_distribution(min_wavelength, max_wavelength, num_wavelengths):
    wavelengths = {}
    wavelengths_list = np.linspace(min_wavelength, max_wavelength, num_wavelengths)
    weight = 1.0 / num_wavelengths  # Equal weight for each wavelength

    for wavelength in wavelengths_list:
        wavelengths[wavelength] = weight

    return wavelengths

def calculate_spot_size(intensity_matrix):
    # Convert the intensity matrix to uint8 type for processing
    intensity_image = np.array(intensity_matrix, dtype=np.uint8)

    # Apply thresholding to extract the spot
    _, thresholded = cv2.threshold(intensity_image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)

    # Find contours of the spot
    contours, _ = cv2.findContours(thresholded, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    # Assuming the largest contour corresponds to the spot
    if len(contours) > 0:
        largest_contour = max(contours, key=cv2.contourArea)
        # Calculate the minimum enclosing circle around the contour
        (_, _), radius = cv2.minEnclosingCircle(largest_contour)
        # Calculate diameter from radius
        spot_diameter = 2 * radius
        return spot_diameter

    return float('inf')

def center_of_energy_mass(matrix):
    total_mass = sum(sum(row) for row in matrix)

    center_row = sum(i * sum(row) for i, row in enumerate(matrix, start=1)) / total_mass
    center_col = sum(j * matrix[i-1][j-1] for i, row in enumerate(matrix, start=1) for j in range(1, len(row)+1)) / total_mass

    return (center_col, center_row)

def encircled_energy(matrix, percent = 0.9):
    center_point = center_of_energy_mass(matrix)
    # Generate a grid of indices
    indices = np.indices(matrix.shape)

    # Calculate distances from each point in the matrix to the center point
    distances = np.sqrt((indices[0] - center_point[0])**2 + (indices[1] - center_point[1])**2)

    # Flatten the distances and spot matrix for sorting
    flat_distances = distances.flatten()
    flat_spot_matrix = matrix.flatten()

    # Get the indices of the distances within the potential circle
    inside_circle_indices = flat_distances <= flat_distances.max()

    # Sort the relevant distances and corresponding spot matrix values
    sorted_indices = np.argsort(flat_distances[inside_circle_indices])
    sorted_distances = flat_distances[inside_circle_indices][sorted_indices]
    sorted_spot_matrix = flat_spot_matrix[inside_circle_indices][sorted_indices]

    # Calculate cumulative distribution of energy
    cumulative_energy = np.cumsum(sorted_spot_matrix)

    # Find the index where cumulative energy exceeds 90%
    threshold_index = np.argmax(cumulative_energy >= percent * cumulative_energy[-1])

    # Use the distance at the threshold index as the radius
    radius = sorted_distances[threshold_index]

    return radius, center_point

def absolute_pixel_size(resolution, size):
    # Calculate pixel size for width and height
    pixel_size_width = size[0] / resolution[0]
    pixel_size_height = size[1] / resolution[1]

    return (pixel_size_width, pixel_size_height)

def visualize_spot(matrix, center_point, radius):
    plt.figure(figsize=(10, 10))
    plt.imshow(matrix)
    plt.scatter(center_point[1], center_point[0], marker='x', color='red')
    plt.gca().add_patch(plt.Circle((center_point[1], center_point[0]), radius, color='red', fill=False))
    plt.show()

def visualize_matrix(matrix, interactive = False, title = None):

    if not interactive:
        plt.figure(figsize=(10, 10))
        plt.imshow(matrix)
        if title:
            plt.title(title)
        plt.show()

    if interactive:
        fig = px.imshow(matrix, color_continuous_scale = v.COLOR_SCALE)
        if title:
            fig.update_layout(title_text = title)
        fig.show()



from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
import subprocess
import os

with open('README.md') as f:
    long_description = f.read()

with open('src/threed_optix/package_utils/vars.py', 'r') as f:
    for line in f:
        if 'VERSION' in line:
            version = line.split('=')[1].strip().strip('"')
            break


setup(
    name='threed_optix',
    version=version,
    license='MIT',
    author="3DOptix",
    author_email='ereztep@3doptix.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://3doptix.com',
    keywords=["Optics", "Optical Design", "Optical Simulation", "Optical Design Software", "3DOptix"],
    install_requires=[
        'requests',
        'pandas',
        'matplotlib',
        'plotly',
        'colorama',
        'nbformat',
        'numpy',
        'scikit-image',
        'scipy',
        'opencv-python',
        'dill',
        "ply",
        "networkx",
        "pytest",
        #'matlab2python @ git+https://github.com/ebranlard/matlab2python.git#egg=matlab2python',
    ],
)



import math
import numpy as np

def is_normalized(vector):
    return np.linalg.norm(vector) == 1

def _3d_distance(a, b):

    a = np.array(a)
    b = np.array(b)

    return np.linalg.norm(a - b)

def _compose(pose):
    rotation = pose[3:]
    position = pose[:3]
    x = rotation[0]
    y = rotation[1]
    z = rotation[2]

    cos = math.cos
    sin = math.sin

    c1 = cos(x/2)
    c2 = cos(y/2)
    c3 = cos(z/2)

    s1 = sin(x/2)
    s2 = sin(y/2)
    s3 = sin(z/2)

    qx = s1 * c2 * c3 + c1 * s2 * s3
    qy = c1 * s2 * c3 - s1 * c2 * s3
    qz = c1 * c2 * s3 + s1 * s2 * c3
    qw = c1 * c2 * c3 - s1 * s2 * s3

    qx2 = (qx * qx)
    qy2 = (qy * qy)
    qz2 = (qz * qz)
    qxqy = (qx * qy)
    qyqz = (qy * qz)
    qzqx = (qz * qx)
    qwqx = (qw * qx)
    qwqy = (qw * qy)
    qwqz = (qw * qz)

    d10 = (1 - 2 * (qy2 + qz2))
    d11 = 2 * (qxqy + qwqz)
    d12 = 2 * (qzqx - qwqy)
    d13 = 0
    d20 = 2 * (qxqy - qwqz)
    d21 = (1 - 2 * (qx2 + qz2))
    d22 = 2 * (qyqz + qwqx)
    d23 = 0
    d30 = 2 * (qzqx + qwqy)
    d31 = 2 * (qyqz - qwqx)
    d32 = (1 - 2 * (qx2 + qy2))
    d33 = 0
    d40 = position[0]
    d41 = position[1]
    d42 = position[2]
    d43 = 1

    d1 = [d10, d11, d12, d13]
    d2 = [d20, d21, d22, d23]
    d3 = [d30, d31, d32, d33]
    d4 = [d40, d41, d42, d43]

    matrix = np.array([d1, d2, d3, d4])

    return matrix

def _matmul(a, b):
    return np.matmul(a, b)

def _invert(a):
    return np.linalg.inv(a)

def rad_to_deg(radians):
    return math.degrees(radians)

def deg_to_rad(degrees):
    return math.radians(degrees)
